from typing import Optional, List
import torch
import torch.nn as nn
import torch.distributed as dist

from .module import ScaledLipschitzModule, SharedLipFactory


class LayerCentering(nn.Module):
    r"""
    Applies Layer centering over a mini-batch of inputs.

    This layer implements the operation as described in
    .. math::
        y = x - \mathrm{E}[x] + \beta
    The mean is calculated over the last `D` dimensions
    given in the `dim` parameter.
    `\beta` is learnable bias parameter. that can be
    applied after the mean subtraction.
    Unlike Layer Normalization, this layer is 1-Lipschitz
    This layer uses statistics computed from input data in
    both training and  evaluation modes.
    This layer is :math:`1`-Lipschitz and should be used

    Args:
        num_features: number of features in the input tensor
        dim: dimensions over which to compute the mean
        (default ``input.mean((-2, -1))`` for a 4D tensor).
        bias: if `True`, adds a learnable bias to the output
        of shape (num_features,). Default: `True`

    Shape:
        - Input: :math:`(N, num_features, *)`
        - Output: :math:`(N, num_features, *)` (same shape as input)

    """

    def __init__(self, num_features: int = 1, dim: tuple = [-2, -1], bias=True):
        super(LayerCentering, self).__init__()
        if bias:
            self.bias = nn.Parameter(torch.zeros((num_features,)), requires_grad=True)
        else:
            self.register_parameter("bias", None)
        self.dim = dim

    def forward(self, x):
        mean = x.mean(dim=self.dim, keepdim=True)
        if self.bias is not None:
            bias_shape = (1, -1) + (1,) * (len(x.shape) - 2)
            return x - mean + self.bias.view(bias_shape)
        else:
            return x - mean


LayerCentering2d = LayerCentering


class ScaleBiasLayer(nn.Module):
    def __init__(
        self,
        scalar=1.0,
        num_features: int = 1,
        bias: bool = True,
    ):
        """
        A PyTorch layer that multiplies the input by a fixed scalar.
        and add a bias. This module is used to export
        Lipschitz normalization layers into vanilla layers.
        Args:
            scalar: The scalar multiplier (non-learnable).
            num_features: number of features in the input tensor
            bias: if `True`, adds a learnable bias to the output
            of shape (size,). Default: `True`
        """
        super(ScaleBiasLayer, self).__init__()
        self.scalar = scalar
        self.num_features = num_features
        if bias:
            self.bias = nn.Parameter(torch.zeros((num_features,)), requires_grad=True)
            self.bias_shape = None
        else:
            self.register_parameter("bias", None)

    def forward(self, x):
        if self.bias is not None:
            if self.bias_shape is None:
                self.bias_shape = (1, -1) + (1,) * (len(x.shape) - 2)
            return x * self.scalar + self.bias.view(self.bias_shape)
        else:
            return x * self.scalar


class BatchLipNorm(nn.Module, ScaledLipschitzModule):
    r"""
    Applies Batch Normalization with a single scaling factor over a 2D, 3D, 4D input.

    .. math::

        y_i = (x_i - \mathrm{E}[x_i])/\sqrt{\max{\mathrm{Var}[x_i]+\epsilon}} + \beta_i

    The mean is calculated per-dimension over the mini-batches and
    other dimensions excepted the feature/channel dimension.
        :math:`\beta` is a learnable parameter vectors
    of num_features `C` (where `C` is the number of features or channels of the input).
    This layer uses statistics computed from input data in
    training mode and  a constant in evaluation mode computed as
    the running mean on training samples.
        This layer is compatible with multi-GPU training (torch.nn.distributed).
    Warning: This layer is not :math:`1`-Lipschitz and could be used with a
    SharedLipFactory to compute the global Lipschitz constant on Sequential models.

    Args:
            num_features: number of features in the input tensor
            dim: dimensions over which to compute the mean
        (default ``input.mean((0, -2, -1))`` for a 4D tensor).
            centering: apply a batch centering (\mathrm{E}[x_i]) if `True`
        otherwise only apply the variance normalization.
            bias: if `True`, adds a learnable bias to the output
        of shape (num_features,). Default: `True`
            normalize: apply variance normalization if `True`
        otherwise only apply centering. Note that BatchCentering class
        is equivalent to BatchLipNorm with `normalize=False`.
        By default `normalize=True`.
            factory (optional): SharedLipFactory to register the scaling factor
            eps: a value added to the denominator for numerical stability.

    Shape:
        - Input: :math:`(N, num_features, *)`
        - Output: :math:`(N, num_features, *)` (same shape as input)

    """

    def __init__(
        self,
        num_features: int = 1,
        dim: Optional[tuple] = None,
        centering: bool = True,
        bias: bool = True,
        normalize: bool = True,
        factory: Optional[SharedLipFactory] = None,
        eps: float = 1e-5,
    ):
        nn.Module.__init__(self)
        ScaledLipschitzModule.__init__(self, scaling=normalize, factory=factory)
        self.num_features = num_features
        self.dim = dim
        self.centering = centering
        self.eps = eps

        # register for saving the local mean on batch
        # running_sum_bmean sum of batch mean (mean needs it)
        self.register_buffer("running_sum_bmean", torch.zeros((num_features,)))
        self.register_buffer("running_num_batches", torch.zeros((1,), dtype=torch.long))
        if bias:
            self.bias = nn.Parameter(torch.zeros((num_features,)), requires_grad=True)
        else:
            self.register_parameter("bias", None)

        self.normalize = normalize
        if self.normalize:
            # register running_sum_bmeansq for saving the sum of mean of square on batch
            self.register_buffer("running_sum_bmeansq", torch.zeros(num_features))

            # registers for storing the total number samples in the epoch
            # Need two buffers to keep the total number when averaging
            # at the end of the epoch
            # the epoch average values will be considered as representing one batch
            #  only for the next epoch
            self.register_buffer("running_mean_sample_per_batches", torch.zeros((1,)))
            self.register_buffer("total_num_samples", torch.zeros((1,)))

        # Cached batch stats (for training-time scaling and mean)
        # these cached values are kept outside the forward function
        # to get access in the Lipschitz computation
        self._batch_mean: Optional[torch.Tensor] = None
        self._batch_meansq: Optional[torch.Tensor] = None
        self._batch_count: Optional[torch.Tensor] = None
        self.local_num_elements: Optional[int] = None

        self._first = True

    def _infer_dim(self, x: torch.Tensor) -> tuple:
        if self.dim is None:
            # Default: reduce over batch + spatial dims, keep channel
            self.dim = (0,) + tuple(range(2, x.dim()))

    @staticmethod
    def _all_reduce_sum_(lt: List[torch.Tensor]):
        if dist.is_available() and dist.is_initialized():
            for t in lt:
                dist.all_reduce(t, op=dist.ReduceOp.SUM)

    def reset_states(self):
        self.running_sum_bmean.zero_()
        self.running_num_batches.zero_()
        if self.normalize:
            self.running_sum_bmeansq.zero_()
            self.running_mean_sample_per_batches.zero_()
            self.total_num_samples.zero_()

    # compute average of running values
    def update_running_values(self):
        """
        In eval mode reduce the running values to their mean
        and set the counter of batches to one
        (previous epoch will have a weight equivalent to one batch)
        """
        if self.running_num_batches > 1:
            self._batch_mean = None  # free some memory
            self._batch_meansq = None
            self._batch_count = None
            self.running_sum_bmean /= self.running_num_batches
            if self.normalize:
                self.running_sum_bmeansq /= self.running_num_batches
                self.total_num_samples = self.running_mean_sample_per_batches.clone()
                self.running_mean_sample_per_batches /= self.running_num_batches
            self.running_num_batches.zero_()
            self.running_num_batches += 1

    def _get_mean_and_update(self, training: bool = False) -> torch.Tensor:
        """retrieve the batch mean in training mode
        update and retrieve the running mean in eval mode
        """
        assert training is False, "use _get_mean only in training mode"
        self.update_running_values()
        return self._get_mean(training=False)

    def _get_mean(self, training: bool = False) -> torch.Tensor:
        """retrieve the batch mean in training mode
        and the running mean in eval mode"""

        if training:
            return self._batch_mean

        # eval mode: return running mean
        # case asking for running mean before a step
        if self.running_num_batches.item() == 0:
            return torch.zeros_like(self.running_sum_bmean)
        # self.running_num_batches is set to 1 after update_running_values
        return self.running_sum_bmean / self.running_num_batches

    def _get_var(self, training: bool = False) -> torch.Tensor:
        """retrieve the batch variance in training mode
        and the running variance in eval mode,
        when variance is too small return 1.0 to avoid division by zero
        """
        cur_mean = self._get_mean(training)
        if cur_mean is None:
            raise RuntimeError(
                f"Mean is None when computing variance. training={training}"
            )
        if not self.normalize:
            return torch.ones_like(cur_mean)
        if training:
            # rectified local variance using local buffers:
            # var_sum is sum_on_batch(x_i^2)/current_sample_num
            # where current_sample_num batch size
            # variance =
            # (var_sum - (cur_mean*cur_mean))*current_sample_num/(current_sample_num-1)
            cur_meansq = self._batch_meansq
            cur_sample_num = self.local_num_elements
            assert (
                cur_sample_num is not None
            ), "current_sample_num should be provided in training mode"
            var_factor = cur_sample_num / (cur_sample_num - 1)
            # num_batches = 1.0
        else:
            if self.running_num_batches.item() == 0:
                return torch.ones_like(self.running_sum_bmeansq)
            # rectified variance using running buffers:
            # var_sum is sum_on_epoch(x_i^2)/total_num_samples where
            # total_num_samples number of samples
            # variance =
            # (var_sum - (cur_mean*cur_mean))*total_num_samples/(total_num_samples-1)
            cur_meansq = self.running_sum_bmeansq / self.running_num_batches
            if self.total_num_samples.item() == 0:  # in case of update = False
                self.total_num_samples = self.running_mean_sample_per_batches
            total_num_samples = self.total_num_samples
            var_factor = total_num_samples / (total_num_samples - 1)
            # num_batches = self.running_num_batches

        var = (cur_meansq - (cur_mean * cur_mean)) * var_factor
        var = torch.where(var < self.eps, torch.ones(var.shape).to(var.device), var)
        return var

    def get_scaling(self, training: bool = False):
        """retrieve the scaling factor based on the maximum of the variance
        this is the main difference with batchnorm that uses per-feature variance
        """
        var = self._get_var(training)
        max_var = var.max()
        return 1.0 / max_var.sqrt()

    def forward(self, x):
        self._infer_dim(x)
        mean_shape = (1, -1) + (1,) * (len(x.shape) - 2)
        self.local_num_elements = x.numel() // x.size(1)  # x[:, 0].numel()

        if self.training:
            # on first batch initalize variables
            if self._first:
                self.reset_states()
                self._first = False
            # compute local mean (on batch of a single GPU)
            if self.centering:
                self._batch_mean = x.mean(dim=self.dim)
            else:
                self._batch_mean = torch.zeros((self.num_features,)).to(x.device)

            # compute local mean square (on batch of a single GPU)
            if self.normalize:
                xsq = x * x
                self._batch_meansq = xsq.mean(dim=self.dim)

            aggregated_mean = self._batch_mean.clone().detach()
            # on a single GPU this value is always 1
            num_batches = torch.zeros_like(self.running_num_batches) + 1
            # for multiGPU aggregate mean and num_batches
            list_tensors = [aggregated_mean, num_batches]
            if self.normalize:
                current_meansq = self._batch_meansq.clone()
                list_tensors.append(current_meansq.detach())
            self._all_reduce_sum_(list_tensors)

            # Accumulate running mean, mean square and num elements over the epoch
            # use aggregated mean and mean square for multi GPU
            with torch.no_grad():
                self.running_sum_bmean += aggregated_mean
                self.running_num_batches += num_batches
                if self.normalize:
                    self.running_sum_bmeansq += current_meansq
                    self.running_mean_sample_per_batches += (
                        self.local_num_elements * num_batches
                    )
            mean = self._get_mean(self.training)
        else:
            mean = self._get_mean_and_update(training=self.training)

        # get scaling factor: shape = (C,)
        # get scaling factor: shape = (1,)
        scale = self.get_scaling(training=self.training).to(x.device)

        if self.bias is not None:
            return (x - mean.view(mean_shape)) * scale + self.bias.view(mean_shape)
        else:
            return (x - mean.view(mean_shape)) * scale

    def vanilla_export(self):
        num_features = self.running_sum_bmean.shape[0]
        bias = -self._get_mean_and_update(training=False).detach()

        scalef = 1.0
        if self.normalize:
            scalef = self.get_scaling(training=False).detach()
            bias *= scalef
        # self.running_sum_bmean.detach() / self.running_num_batches.detach()
        if self.bias is not None:
            bias += self.bias.detach()

        layer = ScaleBiasLayer(scalar=scalef, bias=True, num_features=num_features)
        layer.bias.data = bias
        return layer


class BatchCentering(BatchLipNorm):
    r"""
    Applies Batch Centering  over a 2D, 3D, 4D input.

    .. math::

        y_i = (x_i - \mathrm{E}[x_i]) + \beta_i
    BatchCentering implemented as BatchLipNorm with normalize = False,
    factory=None and centering=True.

    The mean is calculated per-dimension over the mini-batches and
    other dimensions excepted the feature/channel dimension.
        :math:`\beta` is a learnable parameter vectors
    of num_features `C` (where `C` is the number of features or channels of the input).
    This layer uses statistics computed from input data in
    training mode and  a constant in evaluation mode computed as
    the running mean on training samples.
        This layer is compatible with multi-GPU training (torch.nn.distributed).
    This layer is :math:`1`-Lipschitz and should be used

    Args:
        num_features: number of features in the input tensor
        dim: dimensions over which to compute the mean
        (default ``input.mean((0, -2, -1))`` for a 4D tensor).
        bias: if `True`, adds a learnable bias to the output
        of shape (num_features,). Default: `True`

    Shape:
        - Input: :math:`(N, num_features, *)`
        - Output: :math:`(N, num_features, *)` (same shape as input)

    """

    def __init__(
        self,
        num_features: int = 1,
        dim: Optional[tuple] = None,
        bias: bool = True,
    ):
        super().__init__(
            num_features=num_features,
            dim=dim,
            centering=True,  # Batchcentering is centering
            bias=bias,
            normalize=False,  # forces scaling_norm = 1
            factory=None,  # forces scaling_norm = 1
        )


BatchCentering2d = BatchCentering



# -*- coding: utf-8 -*-
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================

import pytest
import numpy as np

from . import utils_framework as uft
from .utils_framework import InvertibleDownSampling, InvertibleUpSampling


def check_downsample(x, y, kernel_size):
    index = 0
    for dx in range(kernel_size):
        for dy in range(kernel_size):
            xx = x[:, :, dx::kernel_size, dy::kernel_size]
            yy = y[:, index :: (kernel_size * kernel_size), :, :]
            np.testing.assert_almost_equal(xx, yy, decimal=6)
            index += 1


@pytest.mark.skipif(
    hasattr(InvertibleDownSampling, "unavailable_class"),
    reason="InvertibleDownSampling not available",
)
def test_invertible_downsample():
    x_np = np.arange(32).reshape(1, 2, 4, 4)
    x = uft.to_NCHW_inv(x_np)
    x = uft.to_tensor(x)
    dw_layer = uft.get_instance_framework(InvertibleDownSampling, {"kernel_size": 2})
    y = dw_layer(x)
    y_np = uft.to_numpy(y)
    y_np = uft.to_NCHW(y_np)
    assert y_np.shape == (1, 8, 2, 2)
    check_downsample(x_np, y_np, 2)

    # 2D input
    x_np = np.random.rand(10, 1, 128, 128)  # torch.rand(10, 1, 128, 128)
    x = uft.to_NCHW_inv(x_np)
    x = uft.to_tensor(x)

    dw_layer = uft.get_instance_framework(InvertibleDownSampling, {"kernel_size": 4})
    y = dw_layer(x)
    y_np = uft.to_numpy(y)
    y_np = uft.to_NCHW(y_np)
    assert y_np.shape == (10, 16, 32, 32)
    check_downsample(x_np, y_np, 4)

    x_np = np.random.rand(10, 4, 64, 64)
    x = uft.to_NCHW_inv(x_np)
    x = uft.to_tensor(x)
    dw_layer = uft.get_instance_framework(InvertibleDownSampling, {"kernel_size": 2})
    y = dw_layer(x)
    y_np = uft.to_numpy(y)
    y_np = uft.to_NCHW(y_np)
    assert y_np.shape == (10, 16, 32, 32)
    check_downsample(x_np, y_np, 2)


@pytest.mark.skipif(
    hasattr(InvertibleUpSampling, "unavailable_class"),
    reason="InvertibleUpSampling not available",
)
def test_invertible_upsample():
    # 2D input
    x_np = np.random.rand(10, 16, 32, 32)
    x = uft.to_NCHW_inv(x_np)
    x = uft.to_tensor(x)
    dw_layer = uft.get_instance_framework(InvertibleUpSampling, {"kernel_size": 4})
    y = dw_layer(x)
    y_np = uft.to_numpy(y)
    y_np = uft.to_NCHW(y_np)
    assert y_np.shape == (10, 1, 128, 128)
    check_downsample(y_np, x_np, 4)

    dw_layer = uft.get_instance_framework(InvertibleUpSampling, {"kernel_size": 2})
    y = dw_layer(x)
    y_np = uft.to_numpy(y)
    y_np = uft.to_NCHW(y_np)
    assert y_np.shape == (10, 4, 64, 64)
    check_downsample(y_np, x_np, 2)


@pytest.mark.skipif(
    hasattr(InvertibleUpSampling, "unavailable_class")
    or hasattr(InvertibleDownSampling, "unavailable_class"),
    reason="InvertibleUpSampling not available",
)
def test_invertible_upsample_downsample():
    x_np = np.random.rand(10, 16, 32, 32)
    x = uft.to_NCHW_inv(x_np)
    x = uft.to_tensor(x)
    up_layer = uft.get_instance_framework(InvertibleUpSampling, {"kernel_size": 4})
    y = up_layer(x)

    dw_layer = uft.get_instance_framework(InvertibleDownSampling, {"kernel_size": 4})
    z = dw_layer(y)
    assert z.shape == x.shape
    np.testing.assert_array_equal(x, z)

    x_np = np.random.rand(10, 1, 128, 128)  # torch.rand(10, 1, 128, 128)
    x = uft.to_NCHW_inv(x_np)
    x = uft.to_tensor(x)

    dw_layer = uft.get_instance_framework(InvertibleDownSampling, {"kernel_size": 4})
    y = dw_layer(x)
    up_layer = uft.get_instance_framework(InvertibleUpSampling, {"kernel_size": 4})
    z = up_layer(y)
    assert z.shape == x.shape
    np.testing.assert_array_equal(x, z)



# -*- coding: utf-8 -*-
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
"""
This module contains equivalents for Model and Sequential. These classes add support
for condensation and vanilla exportation.
"""

import abc
import copy
import warnings
import math
from collections import OrderedDict
from typing import List, Optional, Any

import numpy as np
import torch
import torch.nn as nn
import torch.nn.utils.parametrize as parametrize
from torch.nn import Sequential as TorchSequential


def _is_supported_1lip_layer(layer):
    """Return True if the layer is 1-Lipschitz. Note that in some cases, the layer
    is 1-Lipschitz for specific set of parameters.
    """
    supported_1lip_layers = (
        torch.nn.Softmax,
        torch.nn.Flatten,
        torch.nn.Identity,
        torch.nn.ReLU,
        torch.nn.Sigmoid,
        torch.nn.Tanh,
        torch.nn.Unflatten,
    )
    if isinstance(layer, supported_1lip_layers):
        return True
    elif isinstance(layer, torch.nn.MaxPool2d):
        return layer.kernel_size <= layer.stride
    return False


def _get_vanilla_module(module: nn.Module) -> nn.Module:
    """Return the vanilla torch module corresponding to a lipschitz module.

    Args:
        module (nn.Module): A lipschitz module

    Returns:
        nn.Module: The corresponding vanilla torch module
    """
    if isinstance(module, torch.nn.Conv2d):
        return torch.nn.Conv2d(
            in_channels=module.in_channels,
            out_channels=module.out_channels,
            kernel_size=module.kernel_size,
            stride=module.stride,
            padding=module.padding,
            dilation=module.dilation,
            groups=module.groups,
            bias=module.bias is not None,
            padding_mode=module.padding_mode,
        )
    elif isinstance(module, torch.nn.Linear):
        return torch.nn.Linear(
            in_features=module.in_features,
            out_features=module.out_features,
            bias=module.bias is not None,
        )
    elif isinstance(module, torch.nn.ConvTranspose2d):
        return torch.nn.ConvTranspose2d(
            in_channels=module.in_channels,
            out_channels=module.out_channels,
            kernel_size=module.kernel_size,
            stride=module.stride,
            padding=module.padding,
            output_padding=module.output_padding,
            dilation=module.dilation,
            groups=module.groups,
            bias=module.bias is not None,
            padding_mode=module.padding_mode,
        )
    return module


def vanilla_model(model: nn.Module, in_place=True, new_module=None) -> nn.Module:
    """Convert lipschitz modules into their non-lipschitz counterpart (for
    instance, SpectralConv2d layers become Conv2d layers).

    Warning: This function modifies the model in-place by default.
        Set `in_place=False` for a copy of the model.

    Args:
        model (nn.Module): Lipschitz neural network
        in_place (bool): If True, modify the model in place. If False, return a copy
        new_module (nn.Module): If provided, the function will fill this module
            instead of creating a new one. Only useful in recursion, should be None
            when calling the function.
    """

    def _is_exportable(module):
        return hasattr(module, "vanilla_export") and callable(
            getattr(module, "vanilla_export")
        )

    if in_place is False and new_module is None:
        model.eval()
        new_module = copy.deepcopy(model)
        new_module.eval()
    if in_place is True:
        new_module = model

    for n, module in model.named_children():
        if _is_exportable(module):
            # torchlip modules
            setattr(new_module, n, module.vanilla_export())
        elif parametrize.is_parametrized(module) and isinstance(
            module, (nn.Conv2d, nn.Linear, nn.ConvTranspose2d)
        ):
            # compatibility with orthogonium modules
            nmod = _get_vanilla_module(module)
            nmod.weight.data = module.weight.data.clone()
            if nmod.bias is not None:
                nmod.bias.data = module.bias.data.clone()
            else:
                assert module.bias is None
            setattr(new_module, n, nmod)
        elif len(list(module.children())) > 0:
            # compound module, go inside it
            _ = vanilla_model(
                module, in_place=in_place, new_module=new_module.__getattr__(n)
            )
    return new_module


class SharedLipFactory:
    """
    Factory to share scaling factors between multiple layers.
    Register layers of type ScaledLipschitzModule
    that provide a get_scaling_factor method.
    Provide a method to get the product of all scaling factors.
    This assume that the network is sequential
    """

    def __init__(self):
        self.modules: List["ScaledLipschitzModule"] = []

    def register(self, module: "ScaledLipschitzModule"):
        self.modules.append(module)

    def get_current_product_value(self, training):
        """Retrieve the current product of the scaling factors."""
        if not self.modules:
            return torch.ones(())
        scalings = [m.get_scaling_factor(training=training) for m in self.modules]
        return torch.prod(torch.stack(scalings))


class ScaledLipschitzModule(abc.ABC):
    """
    This class allow to set learnable/fixed lipschitz parameter of a layer.
    args:
        scaling: whether the layer has a scaling factor or not.
        If not the scaling factor will return 1.0
        factory: Optional factory to share scaling factors between multiple layers.
    """

    def __init__(
        self, scaling: bool = False, factory: Optional[SharedLipFactory] = None
    ):
        assert (scaling) or (
            factory is None
        ), "Factory has to be none when scaling is False. "
        # Factory of factors
        self.factory = factory
        self.scaling = scaling
        if self.factory is not None:
            self.factory.register(self)

    """Retrieve the scaling_factor of the layer."""

    def get_scaling_factor(self, training: bool = False) -> torch.Tensor:
        if not self.scaling:
            return torch.ones((1,))
        return self.get_scaling(training=training)

    @abc.abstractmethod
    def get_scaling(self, training: bool = False) -> torch.Tensor:
        """Retrieve the scaling factor of the layer."""
        pass

    @abc.abstractmethod
    def vanilla_export(self):
        """
        Convert this layer to a corresponding vanilla torch layer (when possible).
        Returns:
             A vanilla torch version of this layer.
        """
        pass


class _LipschitzCoefMultiplication(nn.Module):
    """Parametrization module for lipschitz global coefficient multiplication."""

    def __init__(self, coef: float):
        super().__init__()
        self._coef = coef

    def forward(self, weight: torch.Tensor) -> torch.Tensor:
        return self._coef * weight


class LipschitzModule(ScaledLipschitzModule):
    """
    This class allow to set lipschitz factor of a layer. Lipschitz layer must inherit
    this class to allow user to set the lipschitz factor.

    Warning:
         This class only regroup useful functions when developing new Lipschitz layers.
         But it does not ensure any property about the layer. This means that
         inheriting from this class won't ensure anything about the lipschitz constant.
    args:
        coefficient_lip: multiplicative coefficient of the layer
        to form a K-Lipschitz layer.
        factory: from ScaledLipschitzModule,
    """

    # The target coefficient:
    _coefficient_lip: float

    def __init__(
        self,
        coefficient_lip: float = 1.0,
        factory: Optional[SharedLipFactory] = None,
    ):
        scaling = coefficient_lip != 1.0
        super().__init__(scaling=scaling, factory=factory)
        self._coefficient_lip = coefficient_lip

    def apply_lipschitz_factor(self):
        """Multiply the layer weights by a lipschitz factor."""
        if self._coefficient_lip == 1.0:
            return
        parametrize.register_parametrization(
            self, "weight", _LipschitzCoefMultiplication(self._coefficient_lip)
        )

    def get_scaling(self, training=False, update=True):
        return self._coefficient_lip


class Sequential(TorchSequential, LipschitzModule):
    def __init__(
        self,
        *args: Any,
        k_coef_lip: float = 1.0,
    ):
        """
        Equivalent of torch.Sequential but allow to set k-lip factor globally. Also
        support condensation and vanilla exportation.
        For now constant repartition is implemented (each layer
        get n_sqrt(k_lip_factor), where n is the number of layers)
        But in the future other repartition function may be implemented.

        Args:
            layers: list of layers to add to the model.
            name: name of the model, can be None
            k_coef_lip: the Lipschitz coefficient to ensure globally on the model.
        """
        TorchSequential.__init__(self, *args)
        LipschitzModule.__init__(self, k_coef_lip)

        # Force the Lipschitz coefficient:
        n_layers = np.sum(
            [isinstance(layer, LipschitzModule) for layer in self.children()]
        )
        for module in self.children():
            if isinstance(module, LipschitzModule):
                module._coefficient_lip = math.pow(k_coef_lip, 1 / n_layers)
            elif _is_supported_1lip_layer(module) is not True:
                warnings.warn(
                    "Sequential model contains a layer which is not a Lipschitz layer: {}".format(  # noqa: E501
                        module
                    )
                )

    def vanilla_export(self):
        """
        Exports this model to a vanilla torch Sequential.

        This method only works for flat sequential. Lipschitz modules are converted
        using their own `vanilla_export` method while non-Lipschitz modules are simply
        copied using `copy.deepcopy`.

        Returns:
            A Vanilla torch.nn.Sequential model.
        """
        layers = []
        for name, layer in self.named_children():
            if isinstance(layer, LipschitzModule):
                layers.append((name, layer.vanilla_export()))
            else:
                layers.append((name, copy.deepcopy(layer)))
        return TorchSequential(OrderedDict(layers))



# -*- coding: utf-8 -*-
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
import pytest
import os
import math

import numpy as np
from . import utils_framework as uft
from .utils_framework import (
    CategoricalCrossentropy,
    ScaledAvgPool2d,
    ScaledAdaptiveAvgPool2d,
    ScaledL2NormPool2d,
    ScaledAdaptativeL2NormPool2d,
)


def check_serialization(layer_type, layer_params):
    input_shape = (4, 10, 10)
    input_shape = uft.to_framework_channel(input_shape)
    m = uft.generate_k_lip_model(layer_type, layer_params, input_shape=input_shape, k=1)
    if m is None:
        return
    loss, optimizer, _ = uft.compile_model(
        m,
        optimizer=uft.get_instance_framework(uft.SGD, inst_params={"model": m}),
        loss=CategoricalCrossentropy(from_logits=True),
    )
    name = layer_type.__class__.__name__
    path = os.path.join("logs", "pooling", name)
    xnp = np.random.uniform(-10, 10, (255,) + input_shape)
    x = uft.to_tensor(xnp)
    y1 = m(x)
    uft.save_model(m, path)
    m2 = uft.load_model(
        path,
        compile=True,
        layer_type=layer_type,
        layer_params=layer_params,
        input_shape=input_shape,
        k=1,
    )
    y2 = m2(x)
    np.testing.assert_allclose(uft.to_numpy(y1), uft.to_numpy(y2))


@pytest.mark.parametrize(
    "layer_type",
    [
        ScaledAvgPool2d,
        ScaledL2NormPool2d,
    ],
)
@pytest.mark.parametrize(
    "layer_params",
    [
        {"kernel_size": 2},
        {"kernel_size": (5, 5)},
        {"kernel_size": 2, "k_coef_lip": 2.5},
        {"kernel_size": (2, 2), "stride": (2, 2)},
    ],
)
def test_pooling_simple(layer_type, layer_params):
    check_serialization(layer_type, layer_params)


@pytest.mark.parametrize(
    "layer_type",
    [
        ScaledAdaptiveAvgPool2d,
        ScaledAdaptativeL2NormPool2d,
    ],
)
@pytest.mark.parametrize(
    "layer_params",
    [
        {"output_size": (1, 1)},
        {"output_size": (1, 1), "k_coef_lip": 2.5},
    ],
)
def test_pooling_global(layer_type, layer_params):
    check_serialization(layer_type, layer_params)


@pytest.mark.parametrize(
    "layer_type,layer_params",
    [
        (
            ScaledAvgPool2d,
            {"kernel_size": 2},
        ),
        (
            ScaledAvgPool2d,
            {"kernel_size": (5, 5)},
        ),
        (
            ScaledAvgPool2d,
            {"kernel_size": 2, "k_coef_lip": 2.5},
        ),
        (ScaledAvgPool2d, {"kernel_size": (2, 2), "stride": (2, 2)}),
        (
            ScaledL2NormPool2d,
            {"kernel_size": 2},
        ),
        (
            ScaledL2NormPool2d,
            {"kernel_size": (5, 5)},
        ),
        (
            ScaledL2NormPool2d,
            {"kernel_size": 2, "k_coef_lip": 2.5},
        ),
        (ScaledL2NormPool2d, {"kernel_size": (2, 2), "stride": (2, 2)}),
        (ScaledAdaptiveAvgPool2d, {"output_size": (1, 1)}),
        (ScaledAdaptiveAvgPool2d, {"output_size": (1, 1), "k_coef_lip": 2.5}),
        (ScaledAdaptativeL2NormPool2d, {"output_size": (1, 1)}),
        (ScaledAdaptativeL2NormPool2d, {"output_size": (1, 1), "k_coef_lip": 2.5}),
    ],
)
def test_pool_vanilla_export(layer_type, layer_params):
    input_shape = (4, 10, 10)
    input_shape = uft.to_framework_channel(input_shape)
    model = uft.generate_k_lip_model(layer_type, layer_params, input_shape, 1.0)

    # lay = SpectralConvTranspose2d(**kwargs)
    # model = Sequential([lay])
    x = np.random.normal(size=(5,) + input_shape)

    x = uft.to_tensor(x)
    y1 = model(x)

    # Test vanilla export inference comparison
    if uft.vanilla_require_a_copy():
        model2 = uft.generate_k_lip_model(layer_type, layer_params, input_shape, 1.0)
        uft.copy_model_parameters(model, model2)
        vanilla_model = uft.vanillaModel(model2)
    else:
        vanilla_model = uft.vanillaModel(model)  # .vanilla_export()
    y2 = vanilla_model(x)
    np.testing.assert_allclose(uft.to_numpy(y1), uft.to_numpy(y2), atol=1e-6)


@pytest.mark.parametrize(
    "layer_type, layer_params, expected",
    [
        (
            ScaledAvgPool2d,
            {"kernel_size": 2},
            [
                [
                    [
                        [10.0 / math.sqrt(2.0 * 2.0), 10.0 / math.sqrt(2.0 * 2.0)],
                        [10.0 / math.sqrt(2.0 * 2.0), 10.0 / math.sqrt(2.0 * 2.0)],
                    ],
                    [
                        [10.0 / math.sqrt(2.0 * 2.0), -2.0 / math.sqrt(2.0 * 2.0)],
                        [1.0 / math.sqrt(2.0 * 2.0), 9.0 / math.sqrt(2.0 * 2.0)],
                    ],
                ]
            ],
        ),
        (
            ScaledL2NormPool2d,
            {"kernel_size": 2},
            [
                [
                    [
                        [math.sqrt(30.0), math.sqrt(30.0)],
                        [math.sqrt(30.0), math.sqrt(30.0)],
                    ],
                    [
                        [math.sqrt(74.0), math.sqrt(22.0)],
                        [math.sqrt(129.0), math.sqrt(95.0)],
                    ],
                ]
            ],
        ),
        (
            ScaledAdaptiveAvgPool2d,
            {"output_size": (1, 1)},
            [[[[40.0 / math.sqrt(4.0 * 4.0)]], [[18.0 / math.sqrt(4.0 * 4.0)]]]],
        ),
        (
            ScaledAdaptativeL2NormPool2d,
            {"output_size": (1, 1)},
            [[[[math.sqrt(120.0)]], [[math.sqrt(320.0)]]]],
        ),
    ],
)
def test_AvgPooling(layer_type, layer_params, expected):
    pool = uft.get_instance_framework(layer_type, layer_params)
    if pool is None:
        return
    input = [
        [  # input shape (bc,c,h,w) = (1,2,4,4)
            [
                [1.0, 2.0, 3.0, 4.0],
                [3.0, 4.0, 1.0, 2.0],
                [1.0, 2.0, 3.0, 4.0],
                [3.0, 4.0, 1.0, 2.0],
            ],
            [
                [6.0, 2.0, 1.0, -4.0],
                [5.0, -3.0, -1.0, 2.0],
                [10.0, -2.0, 3.0, 9.0],
                [-3.0, -4.0, -1.0, -2.0],
            ],
        ]
    ]
    xnp = np.asarray(input)
    xnp = uft.to_NCHW_inv(xnp)  # move channel if needed (TF)
    x = uft.to_tensor(xnp)
    uft.build_layer(pool, x.shape[1:])

    y = pool(x).numpy()
    y = np.squeeze(y)  # yorch keep dim whereas tf not
    y_tnp = np.asarray(expected)
    y_t = uft.to_NCHW_inv(y_tnp)  # move channel if needed (TF)
    y_t = np.squeeze(y_t)
    np.testing.assert_almost_equal(y, y_t, decimal=5)



# -*- coding: utf-8 -*-
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
import pytest
import os
import numpy as np

from . import utils_framework as uft

from .utils_framework import (
    tLinear,
    Loss,
    KRLoss,
    HingeMarginLoss,
    HKRLoss,
    KRMulticlassLoss,
    HingeMulticlassLoss,
    HKRMulticlassLoss,
    SoftHKRMulticlassLoss,
    LseHKRMulticlassLoss,
    MultiMarginLoss,
    TauCategoricalCrossentropyLoss,
    TauSparseCategoricalCrossentropyLoss,
    TauBinaryCrossentropyLoss,
    CategoricalHingeLoss,
    process_labels_for_multi_gpu,
)

# a tester:
# - un cas hardcodé
# - les dtypes pour y_true
# - shape y_pred
# - labels en [0,1] et en [-1,1]
# - liens entre les losses


def get_gaussian_data(n=500, mean1=1.0, mean2=-1.0):
    x1 = np.random.normal(size=(n, 1), loc=mean1, scale=0.1)
    x2 = np.random.normal(size=(n, 1), loc=mean2, scale=0.1)
    y_pred = np.concatenate([x1, x2], axis=0)
    y_true = np.concatenate([np.ones((n, 1)), np.zeros((n, 1))], axis=0)
    return y_pred, y_true


def check_serialization(nb_class, loss):
    layer_type = tLinear
    layer_params = {"in_features": 10, "out_features": nb_class}
    m = uft.generate_k_lip_model(layer_type, layer_params, input_shape=(10,), k=1)
    assert m is not None
    loss, optimizer, _ = uft.compile_model(
        m,
        optimizer=uft.get_instance_framework(uft.SGD, inst_params={"model": m}),
        loss=loss,
    )
    name = loss.__class__.__name__ if isinstance(loss, Loss) else loss.__name__
    path = os.path.join("logs", "losses", name)

    uft.save_model(m, path)
    m2 = uft.load_model(
        path,
        compile=True,
        layer_type=layer_type,
        layer_params=layer_params,
        input_shape=(10,),
        k=1,
    )
    x = np.random.uniform(size=(255, 10))
    x = uft.to_tensor(x)
    m2(x)


def binary_data(x):
    """Return a Framework float32 tensor of shape [N, 1]
    from a list/np.array of shape [N]"""
    return np.expand_dims(np.array(x, dtype=np.float32), axis=-1)


def one_hot_data(x, n_class):
    """Return a Framework float32 tensor of shape [N, n_class]
    from a list/np.array of shape [N]"""
    return np.eye(n_class)[x]


# global test values
y_true1 = [1.0, 1.0, 1.0, 0.0, 0.0, 0.0]
y_pred1 = [0.5, 1.5, -0.5, -0.5, -1.5, 0.5]

y_true1b = [1.0, 1.0, 1.0, 1.0, 0.0, 0.0, 0.0, 0.0]
y_pred1b = [0.5, 1.1, -0.1, 0.7, -1.3, -0.4, 0.2, -0.9]
y_true2 = one_hot_data([0, 0, 0, 1, 1, 2], 3)
y_pred2 = np.float32(
    [
        [2, 0.2, -0.5],
        [-1, -1.2, 0.3],
        [0.8, 2, 0],
        [0, 1, -0.5],
        [2.4, -0.4, -1.1],
        [-0.1, -1.7, 0.6],
    ]
)

y_true3 = one_hot_data([0, 0, 0, 1, 1, 1, 2, 2], 3)
y_pred3 = np.array(
    [
        [1.5, 0.2, -0.5],
        [1, -1.2, 0.3],
        [0.8, 2, -2],
        [0, 1, -0.5],
        [-0.8, 2, 0],
        [2.4, -0.4, -1.1],
        [-0.1, -1.7, 0.6],
        [1.2, 1.3, 2.5],
    ],
    dtype=np.float32,
)
results_tau_cat = np.float64(
    [
        0.044275,
        0.115109,
        1.243572,
        0.084923,
        0.010887,
        2.802300,
        0.114224,
        0.076357,
    ]
)
y_predgaussian, y_truegaussian = get_gaussian_data(20000)
n_class = 10
n_items = 10000
y_truelabels = np.random.randint(0, n_class, n_items)
y_trueonehot = one_hot_data(np.random.randint(0, n_class, n_items), n_class)
y_predonehot = np.random.normal(size=(n_items, n_class))
# Large random data of size [10000, 1]
num_items4 = 10000
y_true4 = np.random.randint(2, size=num_items4)
y_pred4 = np.random.normal(size=(num_items4,))


# TODO add  HKR tests
@pytest.mark.parametrize(
    "loss_instance, loss_params, y_true_np, y_pred_np, expected_loss,rtol",
    [
        (
            KRLoss,
            {},
            binary_data(y_true1),
            binary_data(y_pred1),
            1.0,
            1e-7,
        ),
        (KRLoss, {}, y_truegaussian, y_predgaussian, 2.0, 1e-3),
        (
            HingeMarginLoss,
            {"min_margin": 2.0},
            binary_data(y_true1),
            binary_data(y_pred1),
            4 / 6,
            1e-7,
        ),
        (
            KRMulticlassLoss,
            {"reduction": "auto"},
            y_true2,
            y_pred2,
            np.float32(761.0 / 1800.0),
            1e-7,
        ),
        (
            HKRMulticlassLoss,
            {"alpha": 5.0, "min_margin": 2.0},
            y_true2,
            y_pred2,
            np.float32(1071.0 / 200.0) * uft.scaleAlpha(5.0),
            1e-7,
        ),
        (
            SoftHKRMulticlassLoss,
            {
                "alpha": 5.0,
                "min_margin": 0.2,
            },  # Warning alpha replaced by alpha/(1+alpha)
            y_true2,
            y_pred2,
            np.float32(1.0897621 * uft.scaleDivAlpha(5.0)),
            1e-5,
        ),
        (
            LseHKRMulticlassLoss,
            {
                "alpha": 5.0,
                "min_margin": 0.2,
            },  # Warning alpha replaced by alpha/(1+alpha)
            y_true2,
            y_pred2,
            np.float32(1.0231503),
            1e-5,
        ),
        (
            TauBinaryCrossentropyLoss,
            {"tau": 2.0},
            binary_data(y_true1),
            binary_data(y_pred1),
            0.279185,
            1e-6,
        ),
    ],
)
def test_loss_generic_value(
    loss_instance, loss_params, y_true_np, y_pred_np, expected_loss, rtol
):
    if hasattr(loss_instance, "unavailable_class"):
        pytest.skip(f"{loss_instance} not implemented")
    loss = uft.get_instance_framework(loss_instance, inst_params=loss_params)
    if loss is None:
        pytest.skip(f"{loss_instance}  with params {loss_params} not implemented")
    y_true, y_pred = uft.to_tensor(y_true_np), uft.to_tensor(y_pred_np)

    loss_val = uft.compute_loss(loss, y_pred, y_true).numpy()
    np.testing.assert_allclose(
        loss_val,
        np.float32(expected_loss),
        rtol=rtol,
        err_msg=f"{loss_instance} loss must be equal to {expected_loss}",
    )


@pytest.mark.parametrize(
    "loss_instance, loss_params, y_true_np, y_pred_np, test_minus1",
    [
        (KRLoss, {}, y_truegaussian, y_predgaussian, True),
        (
            HingeMarginLoss,
            {"min_margin": 2.0},
            binary_data(y_true1),
            binary_data(y_pred1),
            True,
        ),
        (
            KRMulticlassLoss,
            {"reduction": "auto"},
            y_trueonehot,
            y_predonehot,
            True,
        ),  # OK (-1,1) but no more one hot
        (
            HingeMulticlassLoss,
            {"min_margin": 2.0},
            y_trueonehot,
            y_predonehot,
            True,
        ),  # OK (-1,1) but no more one hot
        (
            HKRMulticlassLoss,
            {"alpha": 5.0, "min_margin": 2.0},
            y_trueonehot,
            y_predonehot,
            True,
        ),  # OK (-1,1) but no more one hot
        (
            MultiMarginLoss,
            {"min_margin": 1.0},
            y_trueonehot,
            y_predonehot,
            True,
        ),  # OK (-1,1) but no more one hot
        (
            CategoricalHingeLoss,
            {"min_margin": 1.0},
            y_trueonehot,
            y_predonehot,
            True,
        ),  # OK (-1,1) but no more one hot
        (
            TauCategoricalCrossentropyLoss,
            {"tau": 1.0},
            y_trueonehot,
            y_predonehot,
            False,
        ),
        (
            TauSparseCategoricalCrossentropyLoss,
            {"tau": 1.0},
            y_truelabels,
            y_predonehot,
            False,
        ),
        (
            TauBinaryCrossentropyLoss,
            {"tau": 0.5},
            binary_data(y_true1),
            binary_data(y_pred1),
            True,
        ),
    ],
)
def test_loss_generic_equal(
    loss_instance, loss_params, y_true_np, y_pred_np, test_minus1
):
    if hasattr(loss_instance, "unavailable_class"):
        pytest.skip(f"{loss_instance} not implemented")
    loss = uft.get_instance_framework(loss_instance, inst_params=loss_params)
    if loss is None:
        pytest.skip(f"{loss_instance} with params {loss_params} not implemented")
    y_true, y_pred = uft.to_tensor(y_true_np), uft.to_tensor(y_pred_np)
    loss_val = uft.compute_loss(loss, y_pred, y_true).numpy()
    y_true = uft.to_tensor(y_true_np, dtype=uft.type_int32)
    loss_val_2 = uft.compute_loss(loss, y_pred, y_true).numpy()
    np.testing.assert_equal(
        loss_val_2,
        loss_val,
        err_msg=f"{loss_instance} test failed when y_true has dtype int32",
    )
    if test_minus1:
        y_true2_np = np.where(y_true_np == 1.0, 1.0, -1.0)
        y_true2_t = uft.to_tensor(y_true2_np)
        loss_val_3 = uft.compute_loss(loss, y_pred, y_true2_t).numpy()
        np.testing.assert_equal(
            loss_val_3,
            loss_val,
            err_msg="{loss_instance} test failed when labels are in (1, -1)",
        )

    check_serialization(1, loss)


def test_hkr_loss():
    # assuming KRLoss and hinge have been properly tested
    loss_instance = HKRLoss
    if hasattr(loss_instance, "unavailable_class"):
        pytest.skip(f"{loss_instance} not implemented")
    loss = uft.get_instance_framework(
        loss_instance,
        inst_params={"alpha": 0.5, "min_margin": 2.0},
    )
    if loss is None:
        pytest.skip(f"{loss_instance} with params not implemented")
    check_serialization(1, loss)


def test_softhkrmulticlass_loss():
    loss_instance = SoftHKRMulticlassLoss
    if hasattr(loss_instance, "unavailable_class"):
        pytest.skip(f"{loss_instance} not implemented")
    y_true_np, y_pred_np = y_true2, y_pred2
    expected_loss = np.float32(1.0897621) * uft.scaleDivAlpha(
        5.0
    )  # warning alpha scaled to be in [0,1]
    rtol = 1e-5
    loss = uft.get_instance_framework(
        loss_instance, inst_params={"alpha": 5.0, "min_margin": 0.2}
    )
    if loss is None:
        pytest.skip(f"{loss_instance}   with params  not implemented")
    y_true, y_pred = uft.to_tensor(y_true_np), uft.to_tensor(y_pred_np)
    print("y_true shape", y_true2.shape, "y_pred shape", y_pred2.shape)
    loss_val = uft.compute_loss(loss, y_pred, y_true).numpy()
    np.testing.assert_allclose(loss_val, np.float32(expected_loss), rtol=rtol)

    # moving mean should change and thus loss value
    loss_val_bis = uft.compute_loss(loss, y_pred, y_true).numpy()
    np.testing.assert_allclose(
        loss_val_bis, np.float32(1.0834466) * uft.scaleDivAlpha(5.0), rtol=rtol
    )

    y_true_np, y_pred_np = y_trueonehot, y_predonehot
    loss = uft.get_instance_framework(
        loss_instance, inst_params={"alpha": 5.0, "min_margin": 0.2}
    )
    y_true, y_pred = uft.to_tensor(y_true_np), uft.to_tensor(y_pred_np)
    loss_val = uft.compute_loss(loss, y_pred, y_true).numpy()

    # equality require a new instance
    loss2 = uft.get_instance_framework(
        loss_instance, inst_params={"alpha": 5.0, "min_margin": 0.2}
    )
    y_true = uft.to_tensor(y_true_np, dtype=uft.type_int32)
    loss_val_2 = uft.compute_loss(loss2, y_pred, y_true).numpy()
    np.testing.assert_equal(
        loss_val_2,
        loss_val,
        err_msg=f"{loss_instance} test failed when y_true has dtype int32",
    )

    loss3 = uft.get_instance_framework(
        loss_instance, inst_params={"alpha": 5.0, "min_margin": 0.2}
    )
    y_true2_np = np.where(y_true_np == 1.0, 1.0, -1.0)
    y_true2_t = uft.to_tensor(y_true2_np)
    loss_val_3 = uft.compute_loss(loss3, y_pred, y_true2_t).numpy()
    np.testing.assert_equal(
        loss_val_3,
        loss_val,
        err_msg="{loss_instance} test failed when labels are in (1, -1)",
    )
    check_serialization(1, loss)


def test_lsehkrmulticlass_loss():
    loss_instance = LseHKRMulticlassLoss
    if hasattr(loss_instance, "unavailable_class"):
        pytest.skip(f"{loss_instance} not implemented")
    y_true_np, y_pred_np = y_true2, y_pred2
    expected_loss = np.float32(1.0231503)  # warning alpha scaled to be in [0,1]
    rtol = 1e-5
    loss = uft.get_instance_framework(
        loss_instance, inst_params={"alpha": 5.0, "min_margin": 0.2}
    )
    if loss is None:
        pytest.skip(f"{loss_instance}   with params  not implemented")
    y_true, y_pred = uft.to_tensor(y_true_np), uft.to_tensor(y_pred_np)
    print("y_true shape", y_true.shape, "y_pred shape", y_pred.shape)
    loss_val = uft.compute_loss(loss, y_pred, y_true).numpy()
    np.testing.assert_allclose(loss_val, np.float32(expected_loss), rtol=rtol)

    # loss value should not change
    loss_val_bis = uft.compute_loss(loss, y_pred, y_true).numpy()
    np.testing.assert_allclose(loss_val_bis, np.float32(expected_loss), rtol=rtol)

    y_true_np, y_pred_np = y_trueonehot, y_predonehot
    loss = uft.get_instance_framework(
        loss_instance, inst_params={"alpha": 5.0, "min_margin": 0.2}
    )
    y_true, y_pred = uft.to_tensor(y_true_np), uft.to_tensor(y_pred_np)
    loss_val = uft.compute_loss(loss, y_pred, y_true).numpy()

    # equality with a new instance
    loss2 = uft.get_instance_framework(
        loss_instance, inst_params={"alpha": 5.0, "min_margin": 0.2}
    )
    y_true = uft.to_tensor(y_true_np, dtype=uft.type_int32)
    loss_val_2 = uft.compute_loss(loss2, y_pred, y_true).numpy()
    np.testing.assert_equal(
        loss_val_2,
        loss_val,
        err_msg=f"{loss_instance} test failed when y_true has dtype int32",
    )

    loss3 = uft.get_instance_framework(
        loss_instance, inst_params={"alpha": 5.0, "min_margin": 0.2}
    )
    y_true2_np = np.where(y_true_np == 1.0, 1.0, -1.0)
    y_true2_t = uft.to_tensor(y_true2_np)
    loss_val_3 = uft.compute_loss(loss3, y_pred, y_true2_t).numpy()
    np.testing.assert_equal(
        loss_val_3,
        loss_val,
        err_msg="{loss_instance} test failed when labels are in (1, -1)",
    )

    # loss with temperature should be multiplied by temp if min_margin adapted
    temp = 10.0
    loss4 = uft.get_instance_framework(
        loss_instance,
        inst_params={"alpha": 5.0, "min_margin": 0.2 * temp, "temperature": temp},
    )
    loss_val_4 = uft.compute_loss(loss4, y_pred, y_true).numpy()
    np.testing.assert_allclose(
        loss_val_4,
        loss_val * temp,
        err_msg="{loss_instance} test failed when temperature is not 1",
        rtol=1e-4,
        atol=1e-4,
    )

    check_serialization(1, loss)


@pytest.mark.parametrize(
    "loss_instance, loss_params, y_true_np, y_pred_np, expected_loss, rtol",
    [
        (
            KRLoss,
            {"reduction": "none"},
            binary_data(y_true1b),
            binary_data(y_pred1b),
            [1.0, 2.2, -0.2, 1.4, 2.6, 0.8, -0.4, 1.8],
            1e-7,
        ),
        (
            HingeMarginLoss,
            {"min_margin": 0.7 * 2.0, "reduction": "none"},
            binary_data(y_true1b),
            binary_data(y_pred1b),
            [0.2, 0, 0.8, 0, 0, 0.3, 0.9, 0],
            1e-7,
        ),
        (
            HKRLoss,
            {
                "alpha": 2.5,
                "min_margin": 2.0,
                "reduction": "none",
            },
            binary_data(y_true1b),
            binary_data(y_pred1b),
            np.float64([0.25, -2.2, 2.95, -0.65, -2.6, 0.7, 3.4, -1.55])
            * uft.scaleAlpha(2.5),
            1e-7,
        ),
        (
            TauBinaryCrossentropyLoss,
            {"tau": 0.5, "reduction": "none"},
            binary_data(y_true1b),
            binary_data(y_pred1b),
            [1.15188, 0.91098, 1.43692, 1.06676, 0.84011, 1.19628, 1.48879, 0.98650],
            1e-7,
        ),
        (
            KRMulticlassLoss,
            {"reduction": "none"},
            y_true3,
            y_pred3,
            np.float64([326, 314, 120, 250, 496, -258, 396, 450]) / 225,
            1e-7,
        ),  # OK (-1,1) but no more one hot
        (
            HingeMulticlassLoss,
            {"min_margin": 2.0, "reduction": "none"},
            y_true3,
            y_pred3,
            np.float64([17, 13, 34, 15, 12, 62, 17, 45]) / 30,
            1e-7,
        ),  # OK (-1,1) but no more one hot
        (
            MultiMarginLoss,
            {"min_margin": 0.7, "reduction": "none"},
            y_true3,
            y_pred3,
            np.float64([0, 0, 19, 0, 0, 35, 0, 0]) / 30,
            1e-7,
        ),  # OK (-1,1) but no more one hot
        (
            HKRMulticlassLoss,
            {"alpha": 2.5, "min_margin": 1.0, "reduction": "none"},
            y_true3,
            y_pred3,
            np.float64([-779, -656, 1395, -625, -1609, 4557, -1284, 825])
            * uft.scaleAlpha(2.5)
            / 900,
            1e-7,
        ),  # OK (-1,1) but no more one hot
        (
            CategoricalHingeLoss,
            {"min_margin": 1.1, "reduction": "none"},
            y_true3,
            y_pred3,
            np.float64([0, 0.4, 2.3, 0.1, 0, 3.9, 0.4, 0]),
            1e-7,
        ),  # OK (-1,1) but no more one hot
        (
            TauCategoricalCrossentropyLoss,
            {"tau": 2.0, "reduction": "none"},
            y_true3,
            y_pred3,
            results_tau_cat,
            1e-7,
        ),
        (
            TauSparseCategoricalCrossentropyLoss,
            {"tau": 2.0, "reduction": "none"},
            np.argmax(y_true3, axis=-1),
            y_pred3,
            results_tau_cat,
            1e-7,
        ),
        (
            SoftHKRMulticlassLoss,
            {"alpha": 5.0, "min_margin": 0.2, "reduction": "none"},
            y_true3,
            y_pred3,
            np.float64(
                [
                    -0.10878129,
                    0.12582462,
                    2.2788424,
                    -0.17646402,
                    -0.38242298,
                    3.4523692,
                    -0.19672059,
                    1.1028761,
                ]
            )
            * uft.scaleDivAlpha(5.0),
            1e-7,
        ),
        (
            LseHKRMulticlassLoss,
            {"alpha": 5.0, "min_margin": 0.2, "reduction": "none"},
            y_true3,
            y_pred3,
            np.float64(
                [
                    0.03445601,
                    0.21667106,
                    1.9499999,
                    -0.07889394,
                    -0.2494376,
                    2.9666667,
                    -0.11666449,
                    1.0251629,
                ]
            ),
            1e-7,
        ),
    ],
)
def test_no_reduction_loss_generic(
    loss_instance, loss_params, y_true_np, y_pred_np, expected_loss, rtol
):
    """
    Assert binary losses without reduction. Three losses are tested on hardcoded
    y_true/y_pred of shape [8 elements, 1]: KRLoss, HingeMarginLoss and HKRLoss.
    """
    if hasattr(loss_instance, "unavailable_class"):
        pytest.skip(f"{loss_instance} not implemented")
    loss = uft.get_instance_framework(loss_instance, inst_params=loss_params)
    if loss is None:
        pytest.skip(f"{loss_instance}   with params {loss_params} not implemented")
    y_true, y_pred = uft.to_tensor(y_true_np), uft.to_tensor(y_pred_np)
    loss_val = uft.to_numpy(uft.compute_loss(loss, y_pred, y_true))
    loss_val = np.squeeze(loss_val)
    np.testing.assert_allclose(
        loss_val,
        expected_loss,
        rtol=rtol,
        atol=5e-6,
        err_msg=f"Loss {loss} failed",
    )


segments4 = [
    0,
    num_items4 // 2,
    num_items4 // 2 + num_items4 // 4,
    num_items4 // 2 + num_items4 // 4 + num_items4 // 6,
    None,
]
# segments4 = [0,None]


@pytest.mark.skipif(
    hasattr(process_labels_for_multi_gpu, "unavailable_class"),
    reason="process_labels_for_multi_gpu not available",
)
@pytest.mark.parametrize(
    "loss_instance, loss_params, y_true_np, y_pred_np, segments, expected_loss, rtol",
    [
        (
            KRLoss,
            {"multi_gpu": True, "reduction": "sum"},
            binary_data(y_true1b),
            binary_data(y_pred1b),
            [0, 3, 7, None],
            9.2,
            1e-7,
        ),
        (
            HingeMarginLoss,
            {"min_margin": 0.7 * 2.0, "reduction": "sum"},
            binary_data(y_true1b),
            binary_data(y_pred1b),
            [0, 3, 7, None],
            2.2,
            1e-7,
        ),
        (
            HKRLoss,
            {
                "alpha": 2.5,
                "min_margin": 2.0,
                "reduction": "sum",
                "multi_gpu": True,
            },
            binary_data(y_true1b),
            binary_data(y_pred1b),
            [0, 3, 7, None],
            0.3 * uft.scaleAlpha(2.5),
            1e-7,
        ),
        (
            TauBinaryCrossentropyLoss,
            {"tau": 1.5, "reduction": "sum"},
            binary_data(y_true1b),
            binary_data(y_pred1b),
            [0, 3, 7, None],
            2.19262,
            1e-7,
        ),
        (
            KRLoss,
            {"multi_gpu": True, "reduction": "sum"},
            binary_data(y_true4),
            binary_data(y_pred4),
            segments4,
            None,
            5e-6,
        ),
        (
            HingeMarginLoss,
            {"min_margin": 0.7 * 2.0, "reduction": "sum"},
            binary_data(y_true4),
            binary_data(y_pred4),
            segments4,
            None,
            5e-6,
        ),
        (
            HKRLoss,
            {
                "alpha": 2.5,
                "min_margin": 2.0,
                "reduction": "sum",
                "multi_gpu": True,
            },
            binary_data(y_true4),
            binary_data(y_pred4),
            segments4,
            None,
            5e-6,
        ),
        (
            TauBinaryCrossentropyLoss,
            {"tau": 1.5, "reduction": "sum"},
            binary_data(y_true4),
            binary_data(y_pred4),
            segments4,
            None,
            5e-6,
        ),
        (
            KRMulticlassLoss,
            {"multi_gpu": True, "reduction": "sum"},
            y_true3,
            y_pred3,
            [0, 3, 7, None],
            np.float32(698 / 75),
            1e-7,
        ),  # OK (-1,1) but no more one hot
        (
            HingeMulticlassLoss,
            {"min_margin": 2.0, "reduction": "sum"},
            y_true3,
            y_pred3,
            [0, 3, 7, None],
            np.float32(43 / 6),
            1e-7,
        ),  # OK (-1,1) but no more one hot
        (
            MultiMarginLoss,
            {"min_margin": 0.7, "reduction": "sum"},
            y_true3,
            y_pred3,
            [0, 3, 7, None],
            np.float32(9 / 5),
            1e-7,
        ),  # OK (-1,1) but no more one hot
        (
            HKRMulticlassLoss,
            {"multi_gpu": True, "alpha": 2.5, "min_margin": 1.0, "reduction": "sum"},
            y_true3,
            y_pred3,
            [0, 3, 7, None],
            np.float32(152 / 75) * uft.scaleAlpha(2.5),
            1e-7,
        ),  # OK (-1,1) but no more one hot
        (
            KRMulticlassLoss,
            {"multi_gpu": True, "reduction": "sum"},
            binary_data(y_true4),
            binary_data(y_pred4),
            segments4,
            None,
            5e-6,
        ),  # OK (-1,1) but no more one hot
        (
            HingeMulticlassLoss,
            {"min_margin": 2.0, "reduction": "sum"},
            binary_data(y_true4),
            binary_data(y_pred4),
            segments4,
            None,
            5e-6,
        ),  # OK (-1,1) but no more one hot
        (
            MultiMarginLoss,
            {"min_margin": 0.7, "reduction": "sum"},
            binary_data(y_true4),
            binary_data(y_pred4),
            segments4,
            None,
            5e-6,
        ),  # OK (-1,1) but no more one hot
        (
            HKRMulticlassLoss,
            {"multi_gpu": True, "alpha": 2.5, "min_margin": 1.0, "reduction": "sum"},
            binary_data(y_true4),
            binary_data(y_pred4),
            segments4,
            None,
            5e-6,
        ),  # OK (-1,1) but no more one hot
    ],
)
def test_minibatches_binary_loss_generic(
    loss_instance, loss_params, y_true_np, y_pred_np, segments, expected_loss, rtol
):
    # def test_minibatches_binary_losses(self):
    """
    Assert binary losses with mini-batches, to simulate multi-GPU/TPU. Three losses
    are tested using "sum" reduction: KRLoss, HingeMarginLoss and HKRLoss.
    Assertions are tested with:
    - losses on hardcoded small data of size [8, 1] with full batch and mini-batches
    - losses on large random data with full batch and mini-batches.
    """

    if hasattr(loss_instance, "unavailable_class"):
        pytest.skip(f"{loss_instance} not implemented")

    loss = uft.get_instance_framework(loss_instance, inst_params=loss_params)
    if loss is None:
        pytest.skip(f"{loss_instance}  with params {loss_params} not implemented")
    y_true, y_pred = uft.to_tensor(y_true_np), uft.to_tensor(y_pred_np)
    y_true = process_labels_for_multi_gpu(y_true)
    loss_val = uft.compute_loss(loss, y_pred, y_true).numpy()
    if expected_loss is not None:
        np.testing.assert_allclose(
            loss_val,
            expected_loss,
            rtol=rtol,
            atol=5e-6,
            err_msg=f"Loss {loss} failed",
        )
    loss_val_minibatches = 0
    for i in range(len(segments) - 1):
        loss_val_minibatches += uft.compute_loss(
            loss,
            y_pred[segments[i] : segments[i + 1]],
            y_true[segments[i] : segments[i + 1]],
        )

    np.testing.assert_allclose(
        loss_val,
        loss_val_minibatches,
        rtol=rtol,  # 5e-6,
        atol=5e-6,
        err_msg=f"Loss {loss} failed for hardcoded mini-batches",
    )


@pytest.mark.parametrize(
    "loss_instance, loss_params, rtol",
    [
        (
            KRLoss,
            {"name": "KRLoss none", "reduction": "none"},
            1e-7,
        ),
        (
            HingeMarginLoss,
            {"min_margin": 0.4, "reduction": "none", "name": "hinge none"},
            1e-7,
        ),
        (
            HKRLoss,
            {
                "alpha": 5.2,
                "reduction": "none",
                "name": "HKRLoss none",
            },
            1e-7,
        ),
        (
            KRLoss,
            {"name": "KRLoss auto", "reduction": "auto"},
            1e-7,
        ),
        (
            HingeMarginLoss,
            {"min_margin": 0.4, "reduction": "auto", "name": "hinge auto"},
            1e-7,
        ),
        (
            HKRLoss,
            {
                "alpha": 10.0,
                "reduction": "auto",
                "name": "HKRLoss auto",
            },
            1e-7,
        ),
        (
            KRLoss,
            {
                "multi_gpu": True,
                "name": "KRLoss multi_gpu",
                "reduction": "sum",
            },
            1e-7,
        ),
        (
            HKRLoss,
            {
                "alpha": 3.2,
                "multi_gpu": True,
                "reduction": "sum",
                "name": "HKRLoss multi_gpu",
            },
            1e-7,
        ),
    ],
)
def test_multilabel_loss_generic(loss_instance, loss_params, rtol):
    """
    Assert binary losses with multilabels.
    Three losses are tested (KRLoss, HingeMarginLoss and HKRLoss).
    We compare losses with three separate binary classification and
    the corresponding multilabel problem.
    """
    # Create predictions and labels for 3 binary problems and the concatenated
    # multilabel one.

    if hasattr(loss_instance, "unavailable_class"):
        pytest.skip(f"{loss_instance} not implemented")

    y_pred1_np, y_true1_np = get_gaussian_data(1000)
    y_pred2_np, y_true2_np = get_gaussian_data(1000)
    y_pred3_np, y_true3_np = get_gaussian_data(1000)

    y_pred_np = np.concatenate([y_pred1_np, y_pred2_np, y_pred3_np], axis=-1)
    y_true_np = np.concatenate([y_true1_np, y_true2_np, y_true3_np], axis=-1)

    loss = uft.get_instance_framework(loss_instance, inst_params=loss_params)
    if loss is None:
        pytest.skip(f"{loss_instance}  with params {loss_params} not implemented")

    y_true1, y_pred1 = uft.to_tensor(y_true1_np), uft.to_tensor(y_pred1_np)
    loss_val1 = uft.compute_loss(loss, y_pred1, y_true1).numpy()
    y_true2, y_pred2 = uft.to_tensor(y_true2_np), uft.to_tensor(y_pred2_np)
    loss_val2 = uft.compute_loss(loss, y_pred2, y_true2).numpy()
    y_true3, y_pred3 = uft.to_tensor(y_true3_np), uft.to_tensor(y_pred3_np)
    loss_val3 = uft.compute_loss(loss, y_pred3, y_true3).numpy()
    mean_loss_vals = (loss_val1 + loss_val2 + loss_val3) / 3

    y_true, y_pred = uft.to_tensor(y_true_np), uft.to_tensor(y_pred_np)
    loss_val_multilabel = uft.compute_loss(loss, y_pred, y_true).numpy()

    np.testing.assert_allclose(
        loss_val_multilabel,
        mean_loss_vals,
        rtol=rtol,  # 5e-6,
        atol=1e-4,
        err_msg=f"Loss {loss} failed",
    )



# -*- coding: utf-8 -*-
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================

from typing import Union
import torch
from torch.nn.common_types import _size_1_t, _size_2_t
from ..functional import SymmetricPad


class PadConv1d(torch.nn.Conv1d):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: _size_1_t,
        stride: _size_1_t = 1,
        padding: Union[str, _size_1_t] = 0,
        dilation: _size_1_t = 1,
        groups: int = 1,
        bias: bool = True,
        padding_mode: str = "zeros",
    ):
        """
        This class is a Conv1d Layer with additional padding modes

        Args:
            in_channels (int): Number of channels in the input image
            out_channels (int): Number of channels produced by the convolution
            kernel_size (int or tuple): Size of the convolving kernel
            stride (int or tuple, optional): Stride of the convolution.
            padding (int or tuple, optional): Zero-padding added to both sides of
                the input.
            padding_mode (string, optional): ``'zeros'``, ``'reflect'``,
                ``'replicate'``, ``'symmetric'``  or ``'circular'``.
                Default: ``'zeros'``
            dilation (int or tuple, optional): Spacing between kernel elements.
                Has to be one
            groups (int, optional): Number of blocked connections from input
                channels to output channels. Has to be one
            bias (bool, optional): If ``True``, adds a learnable bias to the
                output.

        This documentation reuse the body of the original torch.nn.Conv1d doc.
        """

        self.old_padding = padding
        self.old_padding_mode = padding_mode
        if padding_mode.lower() == "symmetric":
            padding_mode = "zeros"
            padding = "valid"

        super(PadConv1d, self).__init__(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            bias=bias,
            dilation=dilation,
            groups=groups,
            padding_mode=padding_mode,
        )

        if self.old_padding_mode.lower() == "symmetric":
            self.pad = SymmetricPad(self.old_padding, onedim=True)
        else:
            self.pad = lambda x: x

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        return super(PadConv1d, self).forward(self.pad(input))

    def vanilla_export(self):
        if self.old_padding_mode.lower() == "symmetric":
            next_layer_type = PadConv1d
        else:
            next_layer_type = torch.nn.Conv1d

        layer = next_layer_type(
            in_channels=self.in_channels,
            out_channels=self.out_channels,
            kernel_size=self.kernel_size,
            stride=self.stride,
            padding=self.old_padding,
            dilation=self.dilation,
            groups=self.groups,
            bias=self.bias is not None,
            padding_mode=self.old_padding_mode,
        )
        layer.weight.data = self.weight.detach()
        if self.bias is not None:
            layer.bias.data = self.bias.detach()
        return layer


class PadConv2d(torch.nn.Conv2d):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: _size_2_t,
        stride: _size_2_t = 1,
        padding: Union[str, _size_2_t] = 0,
        dilation: _size_2_t = 1,
        groups: int = 1,
        bias: bool = True,
        padding_mode: str = "zeros",
    ):
        """
        This class is a Conv2d Layer with additional padding modes

        Args:
            in_channels (int): Number of channels in the input image
            out_channels (int): Number of channels produced by the convolution
            kernel_size (int or tuple): Size of the convolving kernel
            stride (int or tuple, optional): Stride of the convolution.
            padding (int or tuple, optional): Zero-padding added to both sides of
                the input.
            padding_mode (string, optional): ``'zeros'``, ``'reflect'``,
                ``'replicate'``, ``'symmetric'``  or ``'circular'``.
                Default: ``'zeros'``
            dilation (int or tuple, optional): Spacing between kernel elements.
                Has to be one
            groups (int, optional): Number of blocked connections from input
                channels to output channels. Has to be one
            bias (bool, optional): If ``True``, adds a learnable bias to the
                output.

        This documentation reuse the body of the original torch.nn.Conv2D doc.
        """

        self.old_padding = padding
        self.old_padding_mode = padding_mode
        if padding_mode.lower() == "symmetric":
            # symmetric padding of one pixel can be replaced by replicate
            if (isinstance(padding, int) and padding <= 1) or (
                isinstance(padding, tuple) and padding[0] <= 1 and padding[1] <= 1
            ):
                self.old_padding_mode = padding_mode = "replicate"
            else:
                padding_mode = "zeros"
                padding = "valid"

        super(PadConv2d, self).__init__(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            bias=bias,
            dilation=dilation,
            groups=groups,
            padding_mode=padding_mode,
        )

        if self.old_padding_mode.lower() == "symmetric":
            self.pad = SymmetricPad(self.old_padding)
        else:
            self.pad = lambda x: x

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        return super(PadConv2d, self).forward(self.pad(input))

    def vanilla_export(self):
        if self.old_padding_mode.lower() == "symmetric":
            next_layer_type = PadConv2d
        else:
            next_layer_type = torch.nn.Conv2d

        layer = next_layer_type(
            in_channels=self.in_channels,
            out_channels=self.out_channels,
            kernel_size=self.kernel_size,
            stride=self.stride,
            padding=self.old_padding,
            dilation=self.dilation,
            groups=self.groups,
            bias=self.bias is not None,
            padding_mode=self.old_padding_mode,
        )
        layer.weight.data = self.weight.detach()
        if self.bias is not None:
            layer.bias.data = self.bias.detach()
        return layer



# -*- coding: utf-8 -*-
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
import torch
import torch.nn as nn
import torch.nn.utils.parametrize as parametrize


class _FrobeniusNorm(nn.Module):
    def __init__(self, disjoint_neurons: bool) -> None:
        super().__init__()
        self.dim_norm = 1 if disjoint_neurons else None

    def forward(self, weight: torch.Tensor) -> torch.Tensor:
        return weight / torch.norm(weight, dim=self.dim_norm, keepdim=True)


def frobenius_norm(
    module: nn.Module, name: str = "weight", disjoint_neurons: bool = True
) -> nn.Module:
    r"""
    Applies Frobenius normalization to a parameter in the given module.

    .. math::
         \mathbf{W} = \dfrac{\mathbf{W}}{\Vert{}\mathbf{W}\Vert{}}

    This is implemented via a hook that applies Frobenius normalization before every
    ``forward()`` call.

    Args:
        module: Containing module.
        name: Name of weight parameter.
        disjoint_neurons: Normalize, independently per neuron or not, the matrix weight.

    Returns:
        The original module with the Frobenius normalization hook.

    Example::

        >>> m = frobenius_norm(nn.Linear(20, 40), name='weight')
        >>> m
        Linear(in_features=20, out_features=40, bias=True)

    """
    parametrize.register_parametrization(module, name, _FrobeniusNorm(disjoint_neurons))
    return module


def remove_frobenius_norm(module: nn.Module, name: str = "weight") -> nn.Module:
    r"""
    Removes the Frobenius normalization reparameterization from a module.

    Args:
        module: Containing module.
        name: Name of weight parameter.

    Example:

        >>> m = frobenius_norm(nn.Linear(20, 40))
        >>> remove_frobenius_norm(m)
    """
    for key, m in module.parametrizations[name]._modules.items():
        if isinstance(m, _FrobeniusNorm):
            if len(module.parametrizations["weight"]) == 1:
                parametrize.remove_parametrizations(module, name)
            else:
                del module.parametrizations[name]._modules[key]



# -*- coding: utf-8 -*-
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
""" """

import warnings
import torch

from .normalizers import bjorck_normalization
from .normalizers import DEFAULT_BETA
from .normalizers import DEFAULT_EPS_SPECTRAL
from .normalizers import DEFAULT_MAXITER_SPECTRAL
from .normalizers import DEFAULT_EPS_BJORCK
from .normalizers import DEFAULT_MAXITER_BJORCK
from .normalizers import spectral_normalization


def spectral_(
    tensor: torch.Tensor,
    eps_spectral: float = DEFAULT_EPS_SPECTRAL,
    maxiter_spectral: int = DEFAULT_MAXITER_SPECTRAL,
):
    r"""
    Apply spectral normalization on the given tensor in-place.

    See also :py:func:`spectral_normalization`.

    .. warning::
        This function is provided for completeness but we recommend using
        :py:func:`torch.nn.init.orthogonal_` instead to obtain a proper (semi)
        orthogonal matrix.

    Args:
        tensor: A 2-dimensional ``torch.Tensor``.
        eps_spectral (float): stopping criterion of iterative power method
        maxiter_spectral (int): maximum number of iterations for the power iteration
    """
    warnings.warn(
        "spectral_ initialization is deprecated, use torch.nn.init.orthogonal_ instead"
    )
    with torch.no_grad():
        tensor.copy_(
            spectral_normalization(
                tensor, None, eps=eps_spectral, maxiter=maxiter_spectral
            )[0]
        )


def bjorck_(
    tensor: torch.Tensor,
    eps_spectral: float = DEFAULT_EPS_SPECTRAL,
    maxiter_spectral: int = DEFAULT_MAXITER_SPECTRAL,
    eps_bjorck=DEFAULT_EPS_BJORCK,
    maxiter_bjorck: int = DEFAULT_MAXITER_BJORCK,
    beta: float = DEFAULT_BETA,
):
    r"""
    Apply Bjorck normalization on the given tensor in-place.

    See also :py:func:`bjorck_normalization`.

    .. warning::
        This function is provided for completeness but we recommend using
        :py:func:`torch.nn.init.orthogonal_` instead to obtain a proper (semi)
        orthogonal matrix.

    Args:
        tensor: A 2-dimensional ``torch.Tensor``.
        eps_spectral (float): stopping criterion of iterative power method
        maxiter_spectral (int): maximum number of iterations for the power iteration
        eps_bjorck (float): stopping criterion in bjorck algorithm
        maxiter_bjorck (int): maximum number of iterations for bjorck algorithm
        beta: Value to use for the :math:`\beta` parameter.
    """
    warnings.warn(
        "bjorck_ initialization is deprecated, use torch.nn.init.orthogonal_ instead"
    )
    with torch.no_grad():
        spectral_tensor = spectral_normalization(
            tensor, None, eps=eps_spectral, maxiter=maxiter_spectral
        )[0]
        tensor.copy_(
            bjorck_normalization(
                spectral_tensor, eps=eps_bjorck, beta=beta, maxiter=maxiter_bjorck
            )
        )



# -*- coding: utf-8 -*-
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
"""
This module contains extra activation functions which respect the Lipschitz constant.
It can be added as a layer, or it can be used in the "activation" params for other
layers.
"""

from typing import Optional

import torch
import torch.nn as nn
import numpy as np

from .. import functional as F
from .module import LipschitzModule


class MaxMin(nn.Module, LipschitzModule):
    r"""
    Applies max-min activation.

    If ``input`` is a tensor of shape :math:`(N, C)` and ``dim`` is
    ``None``, the output can be described as:

    .. math::
        \text{out}(N_i, C_{2j}) = \max(\text{input}(N_i, C_j), 0)\\
        \text{out}(N_i, C_{2j + 1}) = \max(-\text{input}(N_i, C_j), 0)

    where :math:`N` is the batch size and :math:`C` is the size of the
    tensor.

    See also :func:`.functional.max_min`.
    """

    def __init__(self, dim: Optional[int] = None, k_coef_lip: float = 1.0):
        r"""
        Args:
            dim: The dimension to apply max-min. If None, will apply to the
                0th dimension if the shape of input is :math:`(C)` or to the
                first if its :math:`(N, C, *)`.
            k_coef_lip: The lipschitz coefficient to enforce.

        Shape:
            - Input: :math:`(C)` or :math:`(N, C, *)` where :math:`*` means
              any number of additional dimensions.
            - Output: :math:`(2C)` is the input shape was :math:`(C)`, or
              :math:`(N, 2C, *)` if ``dim`` is ``None``, otherwise
              :math:`(N, *, 2C_{dim}, *)` where :math:`C_{dim}` is the
              dimension corresponding to the ``dim`` parameter.

        Note:
            M. Blot, M. Cord, et N. Thome, « Max-min convolutional neural networks
            for image classification », in 2016 IEEE International Conference on Image
            Processing (ICIP), Phoenix, AZ, USA, 2016, p. 3678‑3682.

        """
        nn.Module.__init__(self)
        LipschitzModule.__init__(self, k_coef_lip)
        self._dim = dim

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        return F.max_min(input, self._dim) * self._coefficient_lip

    def vanilla_export(self):
        return self


class GroupSort(nn.Module, LipschitzModule):
    r"""
    Applies group-sort activation.

    The activation works by first reshaping the input to a tensor
    of shape :math:`(N', G)` where :math:`G` is the group size and
    :math:`N'` the number of groups, then sorting each group of
    size :math:`G` and then reshaping to the original input shape.

    See also :func:`.functional.group_sort`.
    """

    def __init__(self, group_size: Optional[int] = None, k_coef_lip: float = 1.0):
        """
        Args:
            group_size: group size used when sorting. When None group size
            is set to input size (fullSort behavior)
            data_format: either channels_first or channels_last
            k_coef_lip: The lipschitz coefficient to enforce.

        Shape:
            - Input: :math:`(N,∗)` where :math:`*` means, any number
               of additional dimensions
            - Output: :math:`(N,*)`, same shape as the input.

        Example:
            >>> m = torch.randn(2, 4)
            tensor([[ 0.2805, -2.0528,  0.6478,  0.5745],
                    [-1.4075,  0.0435, -1.2408,  0.2945]])
            >>> torchlip.GroupSort(4)(m)
            tensor([[-2.0528,  0.2805,  0.5745,  0.6478],
                    [-1.4075, -1.2408,  0.0435,  0.2945]])
        """
        nn.Module.__init__(self)
        LipschitzModule.__init__(self, k_coef_lip)
        self.group_size = group_size

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        return F.group_sort(input, self.group_size) * self._coefficient_lip

    def vanilla_export(self):
        return self


class GroupSort2(GroupSort):
    r"""
    Applies group-sort activation with a group size of 2.

    See :class:`GroupSort` for details.

    See also :func:`.functional.group_sort_2`.
    """

    def __init__(self, k_coef_lip: float = 1.0):
        """
        Args:
            k_coef_lip: The lipschitz coefficient to enforce.
        """
        super().__init__(group_size=2, k_coef_lip=k_coef_lip)


class FullSort(GroupSort):
    r"""
    Applies full-sort activation. This is equivalent to group-sort with
    a group-size equals to the size of the input.

    See :class:`GroupSort` for details.

    See also :func:`.functional.full_sort`.
    """

    def __init__(self, k_coef_lip: float = 1.0):
        """
        Args:
            k_coef_lip: The lipschitz coefficient to enforce.
        """
        super().__init__(group_size=None, k_coef_lip=k_coef_lip)


class LPReLU(nn.PReLU, LipschitzModule):
    r"""
    Applies element-wise PReLU activation with Lipschitz constraint:

    .. math::
        LPReLU(x) = \max(0, x) + a' * \min(0, x)

    or

    .. math::
        LPReLU(x) =
        \text{LipschitzPReLU}(x) =
        \begin{cases}
        x, & \text{ if } x \geq 0 \\
        a' * x, & \text{ otherwise }
        \end{cases}

    where :math:`a' = \max(\min(a, k), -k)`, and :math:`a` is a learnable
    parameter.

    See also :class:`torch.nn.PReLU` and :func:`.functional.lipschitz_prelu`.
    """

    def __init__(
        self, num_parameters: int = 1, init: float = 0.25, k_coef_lip: float = 1.0
    ):
        """
        Args:
            num_parameters: Number of :math:`a` to learn. Although it
            takes an ``int`` as input, ` there are only two legitimate
            values: 1, or the number of channels at input.
            init: The initial value of :math:`a`.
            k_coef_lip: The lipschitz coefficient to enforce.
        """
        nn.PReLU.__init__(self, num_parameters=num_parameters, init=init)
        LipschitzModule.__init__(self, k_coef_lip)

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        return F.lipschitz_prelu(input, self.weight, self._coefficient_lip)

    def vanilla_export(self):
        layer = LPReLU(num_parameters=self.num_parameters)
        layer.weight.data = self.weight.data
        return layer


class HouseHolder(nn.Module, LipschitzModule):
    def __init__(self, channels, k_coef_lip: float = 1.0, theta_initializer=None):
        """
        Householder activation:
        [this review](https://openreview.net/pdf?id=tD7eCtaSkR)
        Adapted from [this repository](https://github.com/singlasahil14/SOC)
        """
        nn.Module.__init__(self)
        LipschitzModule.__init__(self, k_coef_lip)
        assert (channels % 2) == 0
        eff_channels = channels // 2

        if isinstance(theta_initializer, float):
            coef_theta = theta_initializer
        else:
            coef_theta = 0.5 * np.pi
        self.theta = nn.Parameter(
            coef_theta * torch.ones(eff_channels), requires_grad=True
        )
        if theta_initializer is not None:
            if isinstance(theta_initializer, str):
                name2init = {
                    "zeros": torch.nn.init.zeros_,
                    "ones": torch.nn.init.ones_,
                    "normal": torch.nn.init.normal_,
                }
                assert (
                    theta_initializer in name2init
                ), f"Unknown initializer {theta_initializer}"
                name2init[theta_initializer](self.theta)
            elif isinstance(theta_initializer, float):
                pass
            else:
                raise ValueError(f"Unknown initializer {theta_initializer}")

    def forward(self, z, axis=1):
        theta_shape = (1, -1) + (1,) * (len(z.shape) - 2)
        theta = self.theta.view(theta_shape)
        x, y = z.split(z.shape[axis] // 2, axis)
        selector = (x * torch.sin(0.5 * theta)) - (y * torch.cos(0.5 * theta))

        a_2 = x * torch.cos(theta) + y * torch.sin(theta)
        b_2 = x * torch.sin(theta) - y * torch.cos(theta)

        a = x * (selector <= 0) + a_2 * (selector > 0)
        b = y * (selector <= 0) + b_2 * (selector > 0)
        return torch.cat([a, b], dim=axis)

    def vanilla_export(self):
        return self



# -*- coding: utf-8 -*-
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
import os
import pytest

import numpy as np

from . import utils_framework as uft

from .utils_framework import (
    BatchCentering,
    BatchLipNorm,
    LayerCentering,
    SharedLipFactory,
)


def check_serialization(
    layer_type, layer_params, input_shape=(10,), norm_factory=False
):
    if norm_factory:
        factory = SharedLipFactory()
        layer_params["factory"] = factory
    m = uft.generate_k_lip_model(layer_type, layer_params, input_shape=input_shape, k=1)
    if m is None:
        pytest.skip()
    loss, optimizer, _ = uft.compile_model(
        m,
        optimizer=uft.get_instance_framework(uft.SGD, inst_params={"model": m}),
        loss=uft.CategoricalCrossentropy(from_logits=True),
    )
    name = layer_type.__class__.__name__
    path = os.path.join("logs", "normalization", name)
    xnp = np.random.uniform(-10, 10, (255,) + input_shape)
    x = uft.to_tensor(xnp)
    y1 = m(x)
    uft.save_model(m, path)
    m2 = uft.load_model(
        path,
        compile=True,
        layer_type=layer_type,
        layer_params=layer_params,
        input_shape=input_shape,
        k=1,
    )
    y2 = m2(x)
    np.testing.assert_allclose(uft.to_numpy(y1), uft.to_numpy(y2))


@pytest.mark.skipif(
    hasattr(LayerCentering, "unavailable_class"),
    reason="LayerCentering not available",
)
@pytest.mark.parametrize(
    "size, input_shape, bias",
    [
        (4, (3, 4, 8, 8), False),
        (4, (3, 4, 8, 8), True),
    ],
)
def test_LayerCentering(size, input_shape, bias):
    """evaluate layerbatch centering"""
    input_shape = uft.to_framework_channel(input_shape)
    x = np.arange(np.prod(input_shape)).reshape(input_shape)
    bn = uft.get_instance_framework(
        LayerCentering, {"num_features": size, "bias": bias}
    )

    mean_x = np.mean(x, axis=(2, 3))
    mean_shape = (-1, size, 1, 1)
    x = uft.to_tensor(x)
    y = bn(x)
    np.testing.assert_allclose(
        uft.to_numpy(y), x - np.reshape(mean_x, mean_shape), atol=1e-5
    )
    y = bn(2 * x)
    np.testing.assert_allclose(
        uft.to_numpy(y), 2 * x - 2 * np.reshape(mean_x, mean_shape), atol=1e-5
    )  # keep substract batch mean
    bn.eval()
    y = bn(2 * x)
    np.testing.assert_allclose(
        uft.to_numpy(y), 2 * x - 2 * np.reshape(mean_x, mean_shape), atol=1e-5
    )  # eval mode use running_mean


@pytest.mark.skipif(
    hasattr(BatchCentering, "unavailable_class"),
    reason="BatchCentering not available",
)
@pytest.mark.parametrize(
    "size, input_shape, bias",
    [
        (4, (3, 4), False),
        (4, (3, 4), True),
        (4, (3, 4, 8, 8), False),
        (4, (3, 4, 8, 8), True),
    ],
)
def test_BatchCentering(size, input_shape, bias):
    """evaluate layerbatch centering"""
    input_shape = uft.to_framework_channel(input_shape)
    x = np.arange(np.prod(input_shape)).reshape(input_shape)
    bn = uft.get_instance_framework(
        BatchCentering, {"num_features": size, "bias": bias}
    )
    if len(input_shape) == 2:
        mean_x = np.mean(x, axis=0)
        mean_shape = (1, size)
    else:
        mean_x = np.mean(x, axis=(0, 2, 3))
        mean_shape = (1, size, 1, 1)
    x = uft.to_tensor(x)
    y = bn(x)
    np.testing.assert_allclose(bn.running_sum_bmean, mean_x, atol=1e-5)
    np.testing.assert_allclose(
        uft.to_numpy(y), x - np.reshape(mean_x, mean_shape), atol=1e-5
    )
    y = bn(2 * x)
    new_runningmean = (mean_x + 2 * mean_x) / 2.0
    np.testing.assert_allclose(
        bn.running_sum_bmean / bn.running_num_batches, new_runningmean, atol=1e-5
    )
    np.testing.assert_allclose(
        uft.to_numpy(y), 2 * x - 2 * np.reshape(mean_x, mean_shape), atol=1e-5
    )  # keep substract batch mean
    bn.eval()
    y = bn(2 * x)
    np.testing.assert_allclose(
        bn.running_sum_bmean, new_runningmean, atol=1e-5
    )  # eval mode running mean freezed
    np.testing.assert_allclose(
        uft.to_numpy(y), 2 * x - np.reshape(new_runningmean, mean_shape), atol=1e-5
    )  # eval mode use running_mean


@pytest.mark.parametrize(
    "norm_type",
    [LayerCentering, BatchCentering],
)
@pytest.mark.parametrize(
    "size, input_shape, bias",
    [
        (10, (10,), False),
        (10, (10,), True),
        (7, (7, 8, 8), False),
        (7, (7, 8, 8), True),
    ],
)
def test_Normalization_serialization(norm_type, size, input_shape, bias):
    # Check serialization
    if hasattr(norm_type, "unavailable_class"):
        pytest.skip(f"{norm_type} not available")
    check_serialization(
        norm_type,
        layer_params={"num_features": size, "bias": bias},
        input_shape=input_shape,
    )


def linear_generator(batch_size, input_shape: tuple):
    """
    Generate data according to a linear kernel
    Args:
        batch_size: size of each batch
        input_shape: shape of the desired input

    Returns:
        a generator for the data

    """
    input_shape = tuple(input_shape)
    while True:
        # pick random sample in [0, 1] with the input shape
        batch_x = np.array(
            np.random.uniform(-10, 10, (batch_size,) + input_shape), dtype=np.float16
        )
        # apply the k lip linear transformation
        batch_y = batch_x
        yield batch_x, batch_y


@pytest.mark.parametrize(
    "norm_type",
    [LayerCentering, BatchCentering],
)
@pytest.mark.parametrize(
    "size, input_shape, bias",
    [
        (10, (10,), True),
        (7, (7, 8, 8), True),
    ],
)
def test_Normalization_bias(norm_type, size, input_shape, bias):
    if hasattr(norm_type, "unavailable_class"):
        pytest.skip(f"{norm_type} not available")
    m = uft.generate_k_lip_model(
        norm_type,
        layer_params={"num_features": size, "bias": bias},
        input_shape=input_shape,
        k=1,
    )
    if m is None:
        pytest.skip()
    loss, optimizer, _ = uft.compile_model(
        m,
        optimizer=uft.get_instance_framework(uft.SGD, inst_params={"model": m}),
        loss=uft.CategoricalCrossentropy(from_logits=True),
    )
    batch_size = 10
    bb = uft.to_numpy(uft.get_layer_by_index(m, 0).bias)
    np.testing.assert_allclose(bb, np.zeros((size,)), atol=1e-5)

    traind_ds = linear_generator(batch_size, input_shape)
    uft.train(
        traind_ds,
        m,
        loss,
        optimizer,
        2,
        batch_size,
        steps_per_epoch=10,
    )

    bb = uft.to_numpy(uft.get_layer_by_index(m, 0).bias)
    assert np.linalg.norm(bb) != 0.0


@pytest.mark.skipif(
    hasattr(BatchCentering, "unavailable_class"),
    reason="BatchCentering not available",
)
@pytest.mark.parametrize(
    "size, input_shape, bias",
    [
        (4, (3, 4), False),
        (4, (3, 4), True),
        (4, (3, 4, 8, 8), False),
        (4, (3, 4, 8, 8), True),
    ],
)
def test_BatchCentering_runningmean(size, input_shape, bias):
    """evaluate batch centering convergence of running mean"""
    input_shape = uft.to_framework_channel(input_shape)
    # start with 0 to set up running mean to zero
    x = np.zeros(input_shape)
    bn = uft.get_instance_framework(
        BatchCentering, {"num_features": size, "bias": bias}
    )
    x = uft.to_tensor(x)
    y = bn(x)

    np.testing.assert_allclose(bn.running_sum_bmean, 0.0, atol=1e-5)

    x = np.random.normal(0.0, 1.0, input_shape)
    if len(input_shape) == 2:
        mean_x = np.mean(x, axis=0)
    else:
        mean_x = np.mean(x, axis=(0, 2, 3))
    x = uft.to_tensor(x)
    for _ in range(1000):
        y = bn(x)  # noqa: F841

    np.testing.assert_allclose(
        bn.running_sum_bmean / bn.running_num_batches, mean_x * 1000 / 1001, atol=1e-4
    )
    bn.eval()
    y = bn(x)
    # updated running mean used in eval mode
    np.testing.assert_allclose(bn.running_sum_bmean, mean_x * 1000 / 1001, atol=1e-4)
    x_shape = uft.to_numpy(x).shape
    meanr = np.reshape(mean_x * 1000 / 1001, (1, -1) + (1,) * (len(x_shape) - 2))
    np.testing.assert_allclose(uft.to_numpy(y), uft.to_numpy(x) - meanr, atol=1e-4)


@pytest.mark.parametrize(
    "layer_type,layer_params",
    [
        (BatchCentering, {"num_features": 10, "bias": False}),
        (BatchCentering, {"num_features": 10, "bias": True}),
        (BatchLipNorm, {"num_features": 10, "centering": True, "bias": False}),
        (BatchLipNorm, {"num_features": 10, "centering": True, "bias": True}),
        (BatchLipNorm, {"num_features": 10, "centering": False, "bias": False}),
        (BatchLipNorm, {"num_features": 10, "centering": False, "bias": True}),
    ],
)
@pytest.mark.parametrize("input_shape", [(10, 10), (), (10,)])
@pytest.mark.parametrize(
    "norm",
    [False, True],
)
@pytest.mark.parametrize(
    "norm_factory",
    [False, True],
)
def test_batchcenter_vanilla_export(
    layer_type, layer_params, input_shape, norm, norm_factory
):
    input_shape = (layer_params["num_features"],) + input_shape
    input_shape = uft.to_framework_channel(input_shape)
    if layer_type == BatchLipNorm:
        layer_params["normalize"] = norm
        if norm and norm_factory:
            factory = SharedLipFactory()
            layer_params["factory"] = factory
    model = uft.generate_k_lip_model(layer_type, layer_params, input_shape, 1.0)

    x = np.random.normal(size=(5,) + input_shape)

    x = uft.to_tensor(x)
    y1 = model(x)
    for _ in range(1000):
        y = model(x)  # noqa: F841

    model.eval()
    y1 = model(x)
    # Test vanilla export inference comparison
    if uft.vanilla_require_a_copy():
        model2 = uft.generate_k_lip_model(layer_type, layer_params, input_shape, 1.0)
        uft.copy_model_parameters(model, model2)
        vanilla_model = uft.vanillaModel(model2)
    else:
        vanilla_model = uft.vanillaModel(model)  # .vanilla_export()
    y2 = vanilla_model(x)
    assert type(list(vanilla_model.children())[0]) is uft.ScaleBiasLayer
    np.testing.assert_allclose(uft.to_numpy(y1), uft.to_numpy(y2), atol=1e-6)


@pytest.mark.parametrize(
    "size, input_shape, bias, norm",
    [
        (4, (3, 4), False, False),
        (4, (3, 4), False, True),
        (4, (3, 4), True, False),
        (4, (3, 4), True, True),
        (4, (3, 4, 8, 8), False, False),
        (4, (3, 4, 8, 8), False, True),
        (4, (3, 4, 8, 8), True, False),
        (4, (3, 4, 8, 8), True, True),
    ],
)
def test_BatchLipNorm(size, input_shape, bias, norm):
    """evaluate layerbatch centering"""
    # input_shape = uft.to_framework_channel(input_shape)
    xnp = np.arange(np.prod(input_shape)).reshape(input_shape)
    factory = None
    if norm:
        factory = SharedLipFactory()
    bn = BatchLipNorm(
        **{"num_features": size, "bias": bias, "normalize": norm, "factory": factory}
    )
    # bn_mom = bn.momentum
    if len(input_shape) == 2:
        mean_x = np.mean(xnp, axis=0)
        var_x = np.var(xnp, axis=0, ddof=1)
        mean_shape = (1, size)
    else:
        mean_x = np.mean(xnp, axis=(0, 2, 3))
        var_x = np.var(xnp, axis=(0, 2, 3), ddof=1)
        mean_shape = (1, size, 1, 1)
    x = uft.to_tensor(xnp)
    y = bn(x)
    scale_factor = current_scale_factor = 1.0
    if norm:
        scale_factor = 1.0 / np.max(np.sqrt(var_x))
    np.testing.assert_allclose(bn._get_mean(), mean_x, atol=1e-5)
    np.testing.assert_allclose(bn.get_scaling_factor(), scale_factor, atol=1e-5)
    np.testing.assert_allclose(
        uft.to_numpy(y),
        (xnp - np.reshape(mean_x, mean_shape)) * scale_factor,
        atol=1e-5,
    )
    y = bn(2 * x)
    new_runningmean = (
        mean_x + 2 * mean_x
    ) / 2.0  # mean_x * (1 - bn_mom) + 2 * mean_x * bn_mom
    # new_runningvar = (var_x + 4 * var_x)/2.0 #mean_x * (1 - bn_mom)
    # + 2 * mean_x * bn_mom
    conc_xnp = np.concatenate([xnp, 2 * xnp], axis=0)
    if len(input_shape) == 2:
        new_runningvar = np.var(conc_xnp, axis=0, ddof=1)
    else:
        new_runningvar = np.var(conc_xnp, axis=(0, 2, 3), ddof=1)
        # (var_x + 4 * var_x + 0.5*(mean_x-2 * mean_x)**2)/2.0 #mean_x * (1 - bn_mom)
        #  + 2 * mean_x * bn_mom
    if norm:
        current_scale_factor = 1.0 / np.max(np.sqrt(4 * var_x))
        scale_factor = 1.0 / np.max(np.sqrt(new_runningvar))
    np.testing.assert_allclose(bn._get_mean(), new_runningmean, atol=1e-5, rtol=1e-5)
    np.testing.assert_allclose(
        bn.get_scaling_factor(), scale_factor, atol=1e-5, rtol=1e-5
    )
    np.testing.assert_allclose(
        uft.to_numpy(y),
        (2 * xnp - 2 * np.reshape(mean_x, mean_shape)) * current_scale_factor,
        atol=1e-5,
        rtol=1e-5,
    )  # keep substract batch mean
    bn.eval()
    y = bn(2 * x)
    np.testing.assert_allclose(
        bn._get_mean(), new_runningmean, atol=1e-5
    )  # eval mode running mean freezed
    np.testing.assert_allclose(
        bn.get_scaling_factor(), scale_factor, atol=1e-5, rtol=1e-5
    )
    np.testing.assert_allclose(
        uft.to_numpy(y),
        (2 * xnp - np.reshape(new_runningmean, mean_shape)) * scale_factor,
        atol=1e-5,
        rtol=1e-5,
    )  # eval mode use running_mean


@pytest.mark.parametrize(
    "norm_type",
    [
        BatchLipNorm,
    ],
)
@pytest.mark.parametrize(
    "norm",
    [False, True],
)
@pytest.mark.parametrize(
    "size, input_shape, bias",
    [
        (10, (10,), False),
        (10, (10,), True),
        (7, (7, 8, 8), False),
        (7, (7, 8, 8), True),
    ],
)
@pytest.mark.parametrize(
    "norm_factory",
    [False, True],
)
def test_batchlipnorm_serialization(
    norm_type, size, input_shape, bias, norm, norm_factory
):
    # Check serialization
    if hasattr(norm_type, "unavailable_class"):
        pytest.skip(f"{norm_type} not available")
    check_serialization(
        norm_type,
        layer_params={"num_features": size, "normalize": norm, "bias": bias},
        input_shape=input_shape,
        norm_factory=norm_factory and norm,
    )


@pytest.mark.parametrize(
    "norm_type",
    [
        BatchLipNorm,
    ],
)
@pytest.mark.parametrize(
    "size, input_shape, bias",
    [
        (10, (10,), True),
        (7, (7, 8, 8), True),
    ],
)
@pytest.mark.parametrize(
    "norm",
    [False, True],
)
@pytest.mark.parametrize(
    "norm_factory",
    [False, True],
)
def test_BatchLipNorm_bias(norm_type, size, input_shape, bias, norm, norm_factory):
    if hasattr(norm_type, "unavailable_class"):
        pytest.skip(f"{norm_type} not available")
    factory = None
    if norm and norm_factory:
        factory = SharedLipFactory()
    m = uft.generate_k_lip_model(
        norm_type,
        layer_params={
            "num_features": size,
            "bias": bias,
            "normalize": norm,
            "factory": factory,
        },
        input_shape=input_shape,
        k=1,
    )
    if m is None:
        pytest.skip()

    loss, optimizer, _ = uft.compile_model(
        m,
        optimizer=uft.get_instance_framework(uft.SGD, inst_params={"model": m}),
        loss=uft.CategoricalCrossentropy(from_logits=True),
    )
    batch_size = 10

    batch_size = 10
    bb = uft.to_numpy(uft.get_layer_by_index(m, 0).bias)
    sf = uft.to_numpy(uft.get_layer_by_index(m, 0).get_scaling_factor())
    np.testing.assert_allclose(bb, np.zeros((size,)), atol=1e-5)
    np.testing.assert_allclose(sf, 1.0, atol=1e-5)

    traind_ds = linear_generator(batch_size, input_shape)

    uft.train(
        traind_ds,
        m,
        loss,
        optimizer,
        2,
        batch_size,
        steps_per_epoch=10,
    )

    bb = uft.to_numpy(uft.get_layer_by_index(m, 0).bias)
    sf = uft.to_numpy(uft.get_layer_by_index(m, 0).get_scaling_factor())
    assert np.linalg.norm(bb) != 0.0
    if norm:
        assert np.linalg.norm(sf) != 1.0
    else:
        assert np.linalg.norm(sf) == 1.0


@pytest.mark.skipif(
    hasattr(BatchLipNorm, "unavailable_class"),
    reason="BatchLipNorm not available",
)
@pytest.mark.parametrize(
    "size, input_shape, bias",
    [
        (4, (3, 4), False),
        (4, (3, 4), True),
        (4, (3, 4, 8, 8), False),
        (4, (3, 4, 8, 8), True),
        (13, (64, 13, 8, 8), False),
        (13, (64, 13, 8, 8), True),
    ],
)
@pytest.mark.parametrize(
    "norm",
    [False, True],
)
@pytest.mark.parametrize(
    "type_seq",
    [0, 1, 2],
)
@pytest.mark.parametrize(
    "norm_factory",
    [False, True],
)
def test_BatchLipNorm_runningmean(
    size, input_shape, bias, norm, type_seq, norm_factory
):
    """evaluate batch centering convergence of running mean"""
    # input_shape = uft.to_framework_channel(input_shape)
    # start with 0 to set up running mean to zero
    if type_seq == 0:
        xnp = np.zeros(input_shape)
        gt_mean = 0.0
        gt_var = 1.0
        epochs = 2
    elif type_seq >= 1:
        epochs = 20
        xnp = np.random.normal(0.0, 1.0, input_shape)
        if len(input_shape) == 2:
            mean_x = np.mean(xnp, axis=0)
            var_x = np.var(xnp, axis=0, ddof=1)
            num_elem = input_shape[0]
        else:
            mean_x = np.mean(xnp, axis=(0, 2, 3))
            var_x = np.var(xnp, axis=(0, 2, 3), ddof=1)
            num_elem = input_shape[0] * input_shape[2] * input_shape[3]
        if type_seq == 1:
            gt_mean = mean_x
            gt_var = var_x * (num_elem - 1) * epochs / (epochs * num_elem - 1)
        else:
            conc_xnp = np.concatenate([xnp, 3 * xnp], axis=0)
            if len(input_shape) == 2:
                new_runningvar = np.var(conc_xnp, axis=0, ddof=1)
            else:
                new_runningvar = np.var(conc_xnp, axis=(0, 2, 3), ddof=1)
            gt_mean = (mean_x + 3 * mean_x) / 2.0
            print("gt mean_x ", mean_x, "new_runningvar ", new_runningvar)
            gt_var = (
                new_runningvar
                * (2 * num_elem - 1)
                * epochs
                / (2 * epochs * num_elem - 1)
            )

    factory = None
    if norm and norm_factory:
        factory = SharedLipFactory()

    bn = uft.get_instance_framework(
        BatchLipNorm,
        {"num_features": size, "bias": bias, "normalize": norm, "factory": factory},
    )

    x = uft.to_tensor(xnp)

    for _ in range(epochs):
        y = bn(x)  # noqa: F841
        if type_seq == 2:
            y = bn(3 * x)  # noqa: F841

    np.testing.assert_allclose(bn._get_mean(), gt_mean, atol=1e-5)
    # constant value => no scale factor
    if (type_seq == 0) or norm:
        print(
            "get_scaling_factor ",
            bn.get_scaling_factor(False),
            " var_x ",
            gt_var,
            1.0 / np.sqrt(np.max(gt_var)),
        )
        np.testing.assert_allclose(
            bn.get_scaling_factor(), 1.0 / np.sqrt(np.max(gt_var)), atol=1e-5
        )



# -*- coding: utf-8 -*-
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
import numpy as np
import torch
from torch.nn.common_types import _size_2_t
from torch.nn.utils.parametrizations import spectral_norm

from ..utils import bjorck_norm
from ..normalizers import DEFAULT_EPS_BJORCK
from ..normalizers import DEFAULT_EPS_SPECTRAL
from ..utils import frobenius_norm
from ..utils import lconv_norm
from .unconstrained import PadConv1d, PadConv2d
from .module import LipschitzModule


class SpectralConv1d(PadConv1d, LipschitzModule):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: _size_2_t,
        stride: _size_2_t = 1,
        padding: _size_2_t = 0,
        dilation: _size_2_t = 1,
        groups: int = 1,
        bias: bool = True,
        padding_mode: str = "zeros",
        k_coef_lip: float = 1.0,
        eps_spectral: int = DEFAULT_EPS_SPECTRAL,
        eps_bjorck: int = DEFAULT_EPS_BJORCK,
    ):
        """
        This class is a Conv1d Layer constrained such that all singular of it's kernel
        are 1. The computation based on BjorckNormalizer algorithm. As this is not
        enough to ensure 1-Lipschitz a coercive coefficient is applied on the
        output.
        The computation is done in three steps:

        1. reduce the largest singular value to 1, using iterated power method.
        2. increase other singular values to 1, using BjorckNormalizer algorithm.
        3. divide the output by the Lipschitz bound to ensure k-Lipschitz.

        Args:
            in_channels (int): Number of channels in the input image
            out_channels (int): Number of channels produced by the convolution
            kernel_size (int or tuple): Size of the convolving kernel
            stride (int or tuple, optional): Stride of the convolution.
            padding (int or tuple, optional): Zero-padding added to both sides of
                the input.
            padding_mode (string, optional): ``'zeros'``, ``'reflect'``,
                ``'replicate'`` or ``'circular'``. Default: ``'zeros'``
            dilation (int or tuple, optional): Spacing between kernel elements.
            groups (int, optional): Number of blocked connections from input
                channels to output channels.
            bias (bool, optional): If ``True``, adds a learnable bias to the
                output.
            k_coef_lip: Lipschitz constant to ensure.
            eps_spectral: stopping criterion for the iterative power algorithm.
            eps_bjorck: stopping criterion Bjorck algorithm (DEFAULT_EPS_BJORCK),
             to discard Bjorck projection (1-Lip only) set this parameter to None.

        This documentation reuse the body of the original torch.nn.Conv1D doc.
        """

        PadConv1d.__init__(
            self,
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            bias=bias,
            dilation=dilation,
            groups=groups,
            padding_mode=padding_mode,
        )
        LipschitzModule.__init__(self, k_coef_lip)

        torch.nn.init.orthogonal_(self.weight)
        if self.bias is not None:
            self.bias.data.fill_(0.0)

        spectral_norm(
            self,
            name="weight",
            eps=eps_spectral,
        )
        if eps_bjorck is not None:  # discard the bjorck norm if None
            bjorck_norm(self, name="weight", eps=eps_bjorck)
        lconv_norm(self)
        self.apply_lipschitz_factor()

    def vanilla_export(self):
        return PadConv1d.vanilla_export(self)


class SpectralConv2d(PadConv2d, LipschitzModule):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: _size_2_t,
        stride: _size_2_t = 1,
        padding: _size_2_t = 0,
        dilation: _size_2_t = 1,
        groups: int = 1,
        bias: bool = True,
        padding_mode: str = "zeros",
        k_coef_lip: float = 1.0,
        eps_spectral: int = DEFAULT_EPS_SPECTRAL,
        eps_bjorck: int = DEFAULT_EPS_BJORCK,
    ):
        """
        This class is a Conv2d Layer constrained such that all singular of it's kernel
        are 1. The computation based on BjorckNormalizer algorithm. As this is not
        enough to ensure 1-Lipschitz a coercive coefficient is applied on the
        output.
        The computation is done in three steps:

        1. reduce the largest singular value to 1, using iterated power method.
        2. increase other singular values to 1, using BjorckNormalizer algorithm.
        3. divide the output by the Lipschitz bound to ensure k-Lipschitz.

        Args:
            in_channels (int): Number of channels in the input image
            out_channels (int): Number of channels produced by the convolution
            kernel_size (int or tuple): Size of the convolving kernel
            stride (int or tuple, optional): Stride of the convolution.
            padding (int or tuple, optional): Zero-padding added to both sides of
                the input.
            padding_mode (string, optional): ``'zeros'``, ``'reflect'``,
                ``'replicate'``, ``'symmetric'``  or ``'circular'``.
                Default: ``'zeros'``
            dilation (int or tuple, optional): Spacing between kernel elements.
                Has to be one
            groups (int, optional): Number of blocked connections from input
                channels to output channels. Has to be one
            bias (bool, optional): If ``True``, adds a learnable bias to the
                output.
            k_coef_lip: Lipschitz constant to ensure.
            eps_spectral: stopping criterion for the iterative power algorithm.
            eps_bjorck: stopping criterion Bjorck algorithm (DEFAULT_EPS_BJORCK),
             to discard Bjorck projection (1-Lip only) set this parameter to None.

        This documentation reuse the body of the original torch.nn.Conv2D doc.
        """

        PadConv2d.__init__(
            self,
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            bias=bias,
            dilation=dilation,
            groups=groups,
            padding_mode=padding_mode,
        )
        LipschitzModule.__init__(self, k_coef_lip)

        torch.nn.init.orthogonal_(self.weight)
        if self.bias is not None:
            self.bias.data.fill_(0.0)

        spectral_norm(
            self,
            name="weight",
            eps=eps_spectral,
        )
        if eps_bjorck is not None:  # discard the bjorck norm if None
            bjorck_norm(self, name="weight", eps=eps_bjorck)
        lconv_norm(self, name="weight")
        self.apply_lipschitz_factor()

    def vanilla_export(self):
        return PadConv2d.vanilla_export(self)


class FrobeniusConv2d(PadConv2d, LipschitzModule):
    """
    Same as SpectralConv2d but in the case of a single output.
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: _size_2_t,
        stride: _size_2_t = 1,
        padding: _size_2_t = 0,
        dilation: _size_2_t = 1,
        groups: int = 1,
        bias: bool = True,
        padding_mode: str = "zeros",
        k_coef_lip: float = 1.0,
    ):
        if np.prod([stride]) != 1:
            raise RuntimeError("FrobeniusConv2d does not support strides")

        PadConv2d.__init__(
            self,
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            padding_mode=padding_mode,
            bias=bias,
            dilation=dilation,
            groups=groups,
        )
        LipschitzModule.__init__(self, k_coef_lip)

        torch.nn.init.orthogonal_(self.weight)
        if self.bias is not None:
            self.bias.data.fill_(0.0)

        frobenius_norm(self, name="weight", disjoint_neurons=False)
        lconv_norm(self)
        self.apply_lipschitz_factor()

    def vanilla_export(self):
        return PadConv2d.vanilla_export(self)


class SpectralConvTranspose2d(torch.nn.ConvTranspose2d, LipschitzModule):
    r"""Applies a 2D transposed convolution operator over an input image
    such that all singular of it's kernel are 1.
    The computation are the same as for SpectralConv2d layer

    Args:
        in_channels (int): Number of channels in the input image
        out_channels (int): Number of channels produced by the convolution
        kernel_size (int or tuple): Size of the convolving kernel
        stride (int or tuple, optional): Stride of the convolution.
        padding (int or tuple, optional): Zero-padding added to both sides of
            the input.
        output_padding: only 0 or none are supported
        padding_mode (string, optional): ``'zeros'``, ``'reflect'``,
            ``'replicate'`` or ``'circular'``. Default: ``'zeros'``
        dilation (int or tuple, optional): Spacing between kernel elements.
            Has to be one.
        groups (int, optional): Number of blocked connections from input
            channels to output channels. Has to be one.
        bias (bool, optional): If ``True``, adds a learnable bias to the
            output.
        k_coef_lip: Lipschitz constant to ensure.
        eps_spectral: stopping criterion for the iterative power algorithm.
        eps_bjorck: stopping criterion Bjorck algorithm (DEFAULT_EPS_BJORCK),
         to discard Bjorck projection (1-Lip only) set this parameter to None.

    This documentation reuse the body of the original torch.nn.ConvTranspose2d
    doc.
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: _size_2_t,
        stride: _size_2_t = 1,
        padding: _size_2_t = 0,
        output_padding: _size_2_t = 0,
        groups: int = 1,
        bias: bool = True,
        dilation: _size_2_t = 1,
        padding_mode: str = "zeros",
        device=None,
        dtype=None,
        k_coef_lip: float = 1.0,
        eps_spectral: int = DEFAULT_EPS_SPECTRAL,
        eps_bjorck: int = DEFAULT_EPS_BJORCK,
    ) -> None:
        if dilation != 1:
            raise ValueError("SpectralConvTranspose2d does not support dilation rate")
        if output_padding not in [0, None]:
            raise ValueError("SpectralConvTranspose2d only supports output_padding=0")
        torch.nn.ConvTranspose2d.__init__(
            self,
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            output_padding=output_padding,
            groups=groups,
            bias=bias,
            dilation=dilation,
            padding_mode=padding_mode,
            device=device,
            dtype=dtype,
        )
        LipschitzModule.__init__(self, k_coef_lip)

        torch.nn.init.orthogonal_(self.weight)
        if self.bias is not None:
            self.bias.data.fill_(0.0)

        spectral_norm(
            self,
            name="weight",
            eps=eps_spectral,
        )
        if eps_bjorck is not None:  # discard the bjorck norm if None
            bjorck_norm(self, name="weight", eps=eps_bjorck)
        lconv_norm(self, name="weight")
        self.apply_lipschitz_factor()

    def vanilla_export(self):
        layer = torch.nn.ConvTranspose2d(
            in_channels=self.in_channels,
            out_channels=self.out_channels,
            kernel_size=self.kernel_size,
            stride=self.stride,
            padding=self.padding,
            output_padding=self.output_padding,
            dilation=self.dilation,
            groups=self.groups,
            bias=self.bias is not None,
            padding_mode=self.padding_mode,
        )
        layer.weight.data = self.weight.detach()
        if self.bias is not None:
            layer.bias.data = self.bias.detach()
        return layer
